from mtools.mplotqueries.plottypes.base_type import BasePlotType

try:
    from matplotlib.dates import date2num
except ImportError as error:
    raise ImportError("Can't import matplotlib. See "
                      "https://matplotlib.org/users/installing.html "
                      "for instructions on how to install matplotlib."
                      "Error: " + str(error))


class EventPlotType(BasePlotType):

    plot_type_str = 'event'

    def plot_group(self, group, idx, axis):
        x = date2num([logevent.datetime for logevent in self.groups[group]])

        # event plots use axvline
        artists = []
        color, marker = self.color_map(group)

        for i, xcoord in enumerate(x):
            if i == 0:
                artist = axis.axvline(xcoord, linewidth=2, picker=5,
                                      color=color, alpha=0.8, label=group)
            else:
                artist = axis.axvline(xcoord, linewidth=2, picker=5,
                                      color=color, alpha=0.8)
            # add meta-data for picking
            artist._mt_plot_type = self
            artist._mt_group = group
            artist._mt_line_id = i
            artists.append(artist)

        axis.autoscale_view(scaley=False)
        return artists

    def clicked(self, event):
        group = event.artist._mt_group
        line_id = event.artist._mt_line_id
        print(self.groups[group][line_id].line_str)


class RSStatePlotType(EventPlotType):
    """
    RSStatePlotType object.

    This plot type derives from the event plot type (vertical lines), but
    instead of plotting arbitrary events, it will only accept lines that
    indicate a replica set change.

    Those lines either contain the string "is now in state" (for other members)
    or are of the form "[rsMgr] replSet PRIMARY" for own state changes.

    A custom group_by method 'lastword()' groups those lines by their last word
    (which is representative of the new state) and an overloaded color_map()
    method assigns colors to each of those states, similar to the ones used in
    MMS.
    """

    plot_type_str = 'rsstate'

    # force group() to always use lastword method to group by
    # group_by = 'lastword'

    colors = ['m', 'y', 'r', 'g', 'g', 'k', 'b', 'c']
    states = ['PRIMARY', 'SECONDARY', 'DOWN', 'STARTUP', 'STARTUP2',
              'RECOVERING', 'ROLLBACK', 'ARBITER']

    def accept_line(self, logevent):
        """
        Return True on match.

        Only match log lines containing 'is now in state' (reflects other
        node's state changes) or of type "[rsMgr] replSet PRIMARY" (reflects
        own state changes).
        """
        if ("is now in state" in logevent.line_str and
                logevent.split_tokens[-1] in self.states):
            return True

        if ("replSet" in logevent.line_str and
                logevent.thread == "rsMgr" and
                logevent.split_tokens[-1] in self.states):
            return True

        return False

    def group_by(self, logevent):
        """Group by the last token of the log line (PRIMARY, SECONDARY,...)."""
        return logevent.split_tokens[-1]

    @classmethod
    def color_map(cls, group):
        print("Group %s" % group)
        """
        Change default color behavior.

        Map certain states always to the same colors (similar to MMS).
        """
        try:
            state_idx = cls.states.index(group)
        except ValueError:
            # on any unexpected state, return black
            state_idx = 5
        return cls.colors[state_idx], cls.markers[0]



from .base_filter import BaseFilter


class FastFilter(BaseFilter):
    """
    FastFilter class.

    Accept only lines that have a duration that is shorter than the specified
    parameter in ms.
    """

    filterArgs = [
        ('--fast', {'action': 'store', 'nargs': '?', 'default': False,
                    'type': int,
                    'help': ('only output lines with query times shorter '
                             'than FAST ms (default 1000)')})
        ]

    def __init__(self, mlogfilter):
        BaseFilter.__init__(self, mlogfilter)
        if ('fast' in self.mlogfilter.args and
                self.mlogfilter.args['fast'] is not False):
            self.active = True
            if self.mlogfilter.args['fast'] is None:
                self.fastms = 1000
            else:
                self.fastms = self.mlogfilter.args['fast']

    def accept(self, logevent):
        """
        Process line.

        Overwrite BaseFilter.accept() and return True if the provided
        logevent should be accepted (causing output), or False if not.
        """
        if self.active and logevent.duration is not None:
            return logevent.duration <= self.fastms
        return False



#!/usr/bin/env python3

from dateutil.tz import tzutc
from pymongo import ASCENDING, DESCENDING

from mtools.util.input_source import InputSource
from mtools.util.logevent import LogEvent

try:
    try:
        from pymongo import MongoClient as Connection
    except ImportError:
        from pymongo import Connection
    from pymongo.errors import ConnectionFailure, AutoReconnect
except ImportError:
    raise ImportError("Can't import pymongo. "
                      "See http://api.mongodb.org/python/current/ "
                      "for instructions on how to install pymongo.")


class ProfileCollection(InputSource):
    """Wrapper class for input source system.profile collection."""

    datetime_format = "ISODate()"

    def __init__(self, hostname='localhost', port=27017, database='test',
                 collection='system.profile'):
        """
        Constructor for ProfileCollection.

        Takes hostname, port, database and collection as parameters. All are
        optional and have default values.
        """
        # store parameters
        self.hostname = hostname
        self.port = port
        self.database = database
        self.collection = collection
        self.name = "%s.%s" % (database, collection)

        # property variables
        self._start = None
        self._end = None
        self._num_events = None

        self.cursor = None

        # test if database can be reached and collection exists
        try:
            mc = Connection(host=hostname, port=port)
            self.versions = [mc.server_info()[u'version']]
            binary = 'mongos' if mc.is_mongos else 'mongod'
            try:
                self.storage_engine = (mc[database].command('serverStatus')
                                       [u'storageEngine'][u'name'])
            except KeyError:
                self.storage_engine = 'mmapv1'

            self.binary = "%s (running on %s:%i)" % (binary, hostname, port)

        except (ConnectionFailure, AutoReconnect) as e:
            raise SystemExit("&s: Can't connect to database, please check if "
                             "a mongod instance is running on %s:%s."
                             % (e, hostname, port))

        self.coll_handle = mc[database][collection]

        if self.coll_handle.count_documents({}, limit=1) == 0:
            raise SystemExit("can't find any data in %s.%s collection. "
                             "Please check database and collection name."
                             % (database, collection))

    @property
    def start(self):
        """Lazy evaluation of start and end of events."""
        if not self._start:
            self._calculate_bounds()
        return self._start

    @property
    def end(self):
        """:azy evaluation of start and end of events."""
        if not self._end:
            self._calculate_bounds()
        return self._end

    @property
    def num_events(self):
        """Lazy evaluation of the number of events."""
        if not self._num_events:
            self._num_events = self.coll_handle.count_documents({})
        return self._num_events

    def next(self):
        """Make iterators."""
        if not self.cursor:
            self.cursor = self.coll_handle.find().sort([("ts", ASCENDING)])

        doc = self.cursor.next()
        doc['thread'] = self.name
        le = LogEvent(doc)
        return le

    def __iter__(self):
        """Iterate over host object. Return LogEvent obj for each document."""
        self.cursor = self.coll_handle.find().sort([("ts", ASCENDING)])

        for doc in self.cursor:
            doc['thread'] = self.name
            le = LogEvent(doc)
            yield le

    def __len__(self):
        """Return the number of events in the collection."""
        return self.num_events

    def _calculate_bounds(self):
        """Calculate beginning and end of log events."""
        # get start datetime
        first = self.coll_handle.find_one(None, sort=[("ts", ASCENDING)])
        last = self.coll_handle.find_one(None, sort=[("ts", DESCENDING)])

        self._start = first['ts']
        if self._start.tzinfo is None:
            self._start = self._start.replace(tzinfo=tzutc())

        self._end = last['ts']
        if self._end.tzinfo is None:
            self._end = self._end.replace(tzinfo=tzutc())

        return True


if __name__ == '__main__':
    pc = ProfileCollection()

    for event in pc:
        print(event)



#!/usr/bin/env python3

from collections import defaultdict


class LogCodeLine(object):
    """
    LogCodeLine represents a logevent pattern extracted from the source code.

    The pattern is a tuple of constant strings, variables are cut out.
    LogCodeLine stores "matches" of the same log pattern from different source
    files and different versions of the code.

    A match is a tuple of (filename, line number, loglevel, trigger). Matches
    are stored in a dictionary with the git tag version as they key, e.g.
    "r2.2.3".

    The import_l2c_db.py tool extracts all such patterns and creates
    LogCodeLines for each pattern.
    """

    def __init__(self, pattern, pattern_id):
        """Constructor takes a pattern, which is a tuple of strings."""
        self.pattern = pattern
        self.pattern_id = pattern_id
        self.versions = set()
        self.matches = defaultdict(list)

    def addMatch(self, version, filename, lineno, loglevel, trigger):
        """
        Add a match to the LogCodeLine.

        Include the version, filename of the source file, the line number, and
        the loglevel.
        """
        self.versions.add(version)
        self.matches[version].append((filename, lineno, loglevel, trigger))

    def __str__(self):
        """
        String representation of a LogCodeLine.

        Outputs all matches of the pattern.
        """
        s = "%s\n" % (" <var> ".join(self.pattern))
        for version in sorted(self.versions):
            for filename, lineno, loglevel, trigger in self.matches[version]:
                s += (f'''{version:>10}: in {filename}:{lineno}, loglevel {loglevel}, trigger {trigger}\n''')
        return s



import datetime

from dateutil import parser

from mtools.util.logevent import LogEvent

line_ctime_pre24 = ("Sun Aug  3 21:52:05 [initandlisten] db version v2.2.4, "
                    "pdfile version 4.5")
line_ctime = "Sun Aug  3 21:52:05.095 [initandlisten] db version v2.4.5"
line_iso8601_local = ("2013-08-03T21:52:05.095+1000 [initandlisten] db "
                      "version v2.5.2-pre-")
line_iso8601_utc = ("2013-08-03T11:52:05.095Z [initandlisten] db "
                    "version v2.5.2-pre-")
line_getmore = ("Mon Aug  5 20:26:32 [conn9] getmore local.oplog.rs "
                "query: { ts: { $gte: new Date(5908578361554239489) } } "
                "cursorid:1870634279361287923 ntoreturn:0 keyUpdates:0 "
                "numYields: 107 locks(micros) r:85093 nreturned:13551 "
                "reslen:230387 144ms")
line_253_numYields = ("2013-10-21T12:07:27.057+1100 [conn2] query test.docs "
                      "query: { foo: 234333.0 } ntoreturn:0 ntoskip:0 "
                      "keyUpdates:0 numYields:1 locks(micros) r:239078 "
                      "nreturned:0 reslen:20 145ms")
line_246_numYields = ("Mon Oct 21 12:14:21.888 [conn4] query test.docs query: "
                      "{ foo: 23432.0 } ntoreturn:0 ntoskip:0 nscanned:316776 "
                      "keyUpdates:0 numYields: 2405 locks(micros) r:743292 "
                      "nreturned:2 reslen:2116 451ms")
line_26_planSummary = ("2014-04-09T23:25:32.502+0000 [conn3355] query "
                       "toothsome.dimgray.cranberry query: { "
                       "cid: ObjectId('535c8bc39194bf134f7757c0') } "
                       "planSummary: IXSCAN { cid: 1 } ntoreturn:20 "
                       "ntoskip:0 keyUpdates:0 numYields:0 locks(micros) "
                       "r:793 nreturned:0 reslen:20 0ms")
line_pattern_26_a = ("2014-03-18T18:34:30.435+1100 [conn10] query test.new "
                     "query: { a: 1.0 } planSummary: EOF ntoreturn:0 "
                     "ntoskip:0 keyUpdates:0 numYields:0 locks(micros) r:103 "
                     "nreturned:0 reslen:20 0ms")
line_pattern_26_b = ("2014-03-18T18:34:34.360+1100 [conn10] query test.new "
                     "query: { query: { a: 1.0 }, orderby: { b: 1.0 } } "
                     "planSummary: EOF ntoreturn:0 ntoskip:0 keyUpdates:0 "
                     "numYields:0 locks(micros) r:55 nreturned:0 "
                     "reslen:20 0ms")
line_pattern_26_c = ("2014-03-18T18:34:50.777+1100 [conn10] query test.new "
                     "query: { $query: { a: 1.0 }, $orderby: { b: 1.0 } } "
                     "planSummary: EOF ntoreturn:0 ntoskip:0 keyUpdates:0 "
                     "numYields:0 locks(micros) r:60 nreturned:0 "
                     "reslen:20 0ms")
line_command_26_a = ("2014-11-21T18:20:20.263+0800 [conn9] command "
                     "admin.$cmd command: replSetGetStatus "
                     "{ replSetGetStatus: 1.0, forShell: 1.0 } keyUpdates:0 "
                     "numYields:0  reslen:76 0ms")
line_command_26_b = ('2014-11-21T18:20:57.076+0800 [conn9] command test.$cmd '
                     'command: aggregate { aggregate: "mgendata", '
                     'pipeline: [ { $group: { _id: "$bar", '
                     'avg: { $avg: "$foo" } } } ], cursor: {} } keyUpdates:0 '
                     'numYields:0 locks(micros) r:6783 reslen:229 11ms')
line_truncated_24 = ("Wed Jan 28 00:31:16.302 [conn12345] warning: log line "
                     "attempted (26k) over max size(10k), printing beginning "
                     "and end ... getmore MyDB.MyColl query: "
                     "{ foo: ObjectId('123456789012345678901234'), "
                     "foo: { $in: [ 1, 2, 3 ] }, bar: false } "
                     "cursorid:1234567890123456789 ntoreturn:0 keyUpdates:0 "
                     "numYields: 23 locks(micros) r:24715 nreturned:1324 "
                     "reslen:256993 1445ms")
line_fassert = ("***aborting after fassert() failure")
line_empty = ("")
line_new_oplog_query = ('2018-05-01T21:57:45.989+0000 I REPL [replication-0] Scheduled new oplog query Fetcher source: host.name database: local query: { find: "oplog.rs", filter: { ts: { $gte: Timestamp(1525211859, 1) } }, tailable: true, oplogReplay: true, awaitData: true, maxTimeMS: 60000, batchSize: 13981010, term: 1, readConcern: { afterClusterTime: Timestamp(1525211859, 1) } } query metadata: { $replData: 1, $oplogQueryData: 1, $readPreference: { mode: "secondaryPreferred" } } active: 1 findNetworkTimeout: 65000ms getMoreNetworkTimeout: 7500ms shutting down?: 0 first: 1 firstCommandScheduler: RemoteCommandRetryScheduler request: RemoteCommand 16543 -- target:host.name db:local cmd:{ find: "oplog.rs", filter: { ts: { $gte: Timestamp(1525211859, 1) } }, tailable: true, oplogReplay: true, awaitData: true, maxTimeMS: 60000, batchSize: 13981010, term: 1, readConcern: { afterClusterTime: Timestamp(1525211859, 1) } } active: 1 callbackHandle.valid: 1 callbackHandle.cancelled: 0 attempt: 1 retryPolicy: RetryPolicyImpl maxAttempts: 1 maxTimeMillis: -1ms')

# fake system.profile documents
profile_doc1 = {"op": "query", "ns": "test.foo",
                "thread": "test.system.profile", "query": {"test": 1},
                "ntoreturn": 0, "ntoskip": 0, "nscanned": 0,
                "keyUpdates": 0, "numYield": 0,
                "lockStats": {"timeLockedMicros": {"r": 461, "w": 0},
                              "timeAcquiringMicros": {"r": 4, "w": 3}},
                "nreturned": 0, "responseLength": 20, "millis": 0,
                "ts": parser.parse("2014-03-20T04:04:21.231Z"),
                "client": "127.0.0.1", "allUsers": [], "user": ""}
profile_doc2 = {"op": "query", "ns": "test.foo",
                "thread": "test.system.profile",
                "query": {"query": {"test": 1},
                          "orderby": {"field": 1}}, "ntoreturn": 0,
                "ntoskip": 0, "nscanned": 0, "keyUpdates": 0,
                "numYield": 0,
                "lockStats": {"timeLockedMicros": {"r": 534, "w": 0},
                              "timeAcquiringMicros": {"r": 5, "w": 4}},
                "nreturned": 0, "responseLength": 20, "millis": 0,
                "ts": parser.parse("2014-03-20T04:04:33.775Z"),
                "client": "127.0.0.1", "allUsers": [], "user": ""}
profile_doc3 = {"op": "query", "ns": "test.foo",
                "thread": "test.system.profile",
                "query": {"$query": {"test": 1},
                          "$orderby": {"field": 1}}, "ntoreturn": 0,
                "ntoskip": 0, "nscanned": 0, "keyUpdates": 0,
                "numYield": 0,
                "lockStats": {"timeLockedMicros": {"r": 436, "w": 0},
                              "timeAcquiringMicros": {"r": 5, "w": 8}},
                "nreturned": 0, "responseLength": 20, "millis": 0,
                "ts": parser.parse("2014-03-20T04:04:52.791Z"),
                "client": "127.0.0.1", "allUsers": [], "user": ""}


def test_logevent_datetime_parsing():
    """Check that all four timestamp formats are correctly parsed."""

    le = LogEvent(line_ctime_pre24)
    this_year = datetime.datetime.now().year

    le_str = le.line_str
    assert(str(le.datetime) == '%s-08-03 21:52:05+00:00' % this_year)
    assert(le._datetime_format == 'ctime-pre2.4')
    assert(le.line_str[4:] == le_str[4:])
    # make sure all datetime objects are timezone aware
    assert(le.datetime.tzinfo is not None)

    le = LogEvent(line_ctime)
    le_str = le.line_str
    assert(str(le.datetime) == '%s-08-03 21:52:05.095000+00:00' % this_year)
    assert(le._datetime_format == 'ctime')
    assert(le.line_str[4:] == le_str[4:])
    # make sure all datetime objects are timezone aware
    assert(le.datetime.tzinfo is not None)

    le = LogEvent(line_iso8601_utc)
    le_str = le.line_str
    assert(str(le.datetime) == '2013-08-03 11:52:05.095000+00:00')
    assert(le._datetime_format == 'iso8601-utc')
    assert(le.line_str[4:] == le_str[4:])
    # make sure all datetime objects are timezone aware
    assert(le.datetime.tzinfo is not None)

    le = LogEvent(line_iso8601_local)
    le_str = le.line_str
    assert(str(le.datetime) == '2013-08-03 21:52:05.095000+10:00')
    assert(le._datetime_format == 'iso8601-local')
    assert(le.line_str[4:] == le_str[4:])
    # make sure all datetime objects are timezone aware
    assert(le.datetime.tzinfo is not None)


def test_logevent_pattern_parsing():
    le = LogEvent(line_pattern_26_a)
    assert(le.pattern) == '{"a": 1}'

    le = LogEvent(line_pattern_26_b)
    assert(le.pattern) == '{"a": 1}'

    le = LogEvent(line_pattern_26_c)
    assert(le.pattern) == '{"a": 1}'


def test_logevent_actual_query_parsing():
    le = LogEvent(line_pattern_26_a)
    assert(le.actual_query) == '{ a: 1.0 }'

    le = LogEvent(line_pattern_26_b)
    assert(le.actual_query) == '{ a: 1.0 }'

    le = LogEvent(line_pattern_26_c)
    assert(le.actual_query) == '{ a: 1.0 }'


def test_logevent_command_parsing():
    le = LogEvent(line_command_26_a)
    assert(le.command) == 'replsetgetstatus'

    le = LogEvent(line_command_26_b)
    assert(le.command) == 'aggregate'

    le = LogEvent(line_getmore)
    assert(le.command) is None


def test_logevent_sort_pattern_parsing():
    le = LogEvent(line_pattern_26_a)
    assert(le.sort_pattern) is None

    le = LogEvent(line_pattern_26_b)
    assert(le.sort_pattern) == '{"b": 1}'

    le = LogEvent(line_pattern_26_c)
    assert(le.sort_pattern) == '{"b": 1}'


def test_logevent_actual_sort_parsing():
    le = LogEvent(line_pattern_26_a)
    assert(le.actual_sort) is None

    le = LogEvent(line_pattern_26_b)
    assert(le.actual_sort) == '{ b: 1.0 }'

    le = LogEvent(line_pattern_26_c)
    assert(le.actual_sort) == '{ b: 1.0 }'


def test_logevent_profile_pattern_parsing():
    le = LogEvent(profile_doc1)
    assert(le.pattern == '{"test": 1}')

    le = LogEvent(profile_doc2)
    assert(le.pattern == '{"test": 1}')

    le = LogEvent(profile_doc3)
    assert(le.pattern == '{"test": 1}')


def test_logevent_profile_sort_pattern_parsing():
    le = LogEvent(profile_doc1)
    assert(le.sort_pattern is None)

    le = LogEvent(profile_doc2)
    assert(le.sort_pattern == '{"field": 1}')

    le = LogEvent(profile_doc3)
    assert(le.sort_pattern == '{"field": 1}')


def test_logevent_extract_new_and_old_numYields():
    le = LogEvent(line_246_numYields)
    assert(le.numYields == 2405)

    le = LogEvent(line_253_numYields)
    assert(le.numYields == 1)


def test_logevent_parse_truncated_line():
    le = LogEvent(line_truncated_24)
    assert(le.thread == "conn12345")
    assert(le.numYields == 23)
    assert(le.operation == "getmore")


def test_logevent_extract_planSummary():
    le = LogEvent(line_26_planSummary)
    assert(le.planSummary == "IXSCAN")

    le = LogEvent(line_pattern_26_a)
    assert(le.planSummary == "EOF")


def test_logevent_extract_actualPlanSummary():
    le = LogEvent(line_26_planSummary)
    assert(le.actualPlanSummary == "IXSCAN { cid: 1 }")

    le = LogEvent(line_pattern_26_a)
    assert(le.actualPlanSummary == "EOF")


def test_logevent_value_extraction():
    """ Check for correct value extraction of all fields. """

    le = LogEvent(line_getmore)
    assert(le.thread == 'conn9')
    assert(le.operation == 'getmore')
    assert(le.namespace == 'local.oplog.rs')
    assert(le.duration == 144)
    assert(le.numYields == 107)
    assert(le.r == 85093)
    assert(le.ntoreturn == 0)
    assert(le.nreturned == 13551)
    assert(le.pattern == '{"ts": 1}')


def test_logevent_non_log_line():
    """ Check that LogEvent correctly ignores non log lines"""
    le = LogEvent(line_fassert)
    assert(le.thread == None)
    assert(le.operation == None)
    assert(le.namespace == None)
    assert(le.duration == None)
    assert(le.numYields == None)
    assert(le.r == None)
    assert(le.ntoreturn == None)
    assert(le.nreturned == None)
    assert(le.pattern == None)

    le = LogEvent(line_empty)
    assert(le.thread == None)
    assert(le.operation == None)
    assert(le.namespace == None)
    assert(le.duration == None)
    assert(le.numYields == None)
    assert(le.r == None)
    assert(le.ntoreturn == None)
    assert(le.nreturned == None)
    assert(le.pattern == None)


def test_logevent_new_oplog_query():
    """ Check that LogEvent correctly ignores new oplog query for duration extraction """
    le = LogEvent(line_new_oplog_query)
    assert(le.duration == None)


def test_logevent_lazy_evaluation():
    """ Check that all LogEvent variables are evaluated lazily. """

    fields = ['_thread', '_operation', '_namespace', '_duration',
              '_numYields', '_r', '_ntoreturn', '_nreturned', '_pattern']

    # before parsing all member variables need to be None
    le = LogEvent(line_getmore)
    for attr in fields:
        assert(getattr(le, attr) is None)

    # after parsing, they all need to be filled out
    le.parse_all()
    for attr in fields:
        assert(getattr(le, attr) is not None)



from datetime import MAXYEAR, datetime, timedelta

from dateutil.tz import tzutc

from .base_filter import BaseFilter
from mtools.util import OrderedDict
from mtools.util.hci import DateTimeBoundaries


def custom_parse_dt(value):
    return value


class DateTimeFilter(BaseFilter):
    """
    DateTimeFilter class.

    This filter has two parser arguments: --from and --to, both are
    optional. All possible values for --from and --to can be described as:

    [DATE] [TIME] [OFFSET] in that order, separated by a space.

    [DATE] can be any of
        - a 3-letter weekday (Mon, Tue, Wed, ...)
        - a date as 3-letter month, 1-2 digits day (Sep 5, Jan 31, Aug 08)
        - the words: today, now, start, end

    [TIME] can be any of
        - hours and minutes (20:15, 04:00, 3:00)
        - hours, minutes and seconds (13:30:01, 4:55:55)

    [OFFSET] consists of [OPERATOR][VALUE][UNIT]   (no spaces in between)

    [OPERATOR] can be + or - (note that - can only be used if the whole
        "[DATE] [TIME] [OFFSET]" is in quotation marks, otherwise it would
        be confused with a separate parameter)

    [VALUE] can be any number

    [UNIT] can be any of s, sec, m, min, h, hours, d, days, w, weeks, mo,
        months, y, years

    The [OFFSET] is added/subtracted to/from the specified [DATE] [TIME].

    For the --from parameter, the default is the same as 'start'
        (0001-01-01 00:00:00). If _only_ an [OFFSET] is given, it is
        added to 'start' (which is not very useful).

    For the --to parameter, the default is the same as 'end'
        (9999-31-12 23:59:59). If _only_ an [OFFSET] is given, however,
        it is added to [FROM].

    Examples:
        --from Sun 10:00
            goes from last Sunday 10:00:00am to the end of the file

        --from Sep 29
            goes from Sep 29 00:00:00 to the end of the file

        --to today 15:00
            goes from the beginning of the file to today at 15:00:00

        --from today --to +1h
            goes from today's date 00:00:00 to today's date 01:00:00

        --from 20:15 --to +3m
            goes from today's date at 20:15:00 to today's date at 20:18:00
    """

    filterArgs = [
        ('--from', {'action': 'store', 'type': custom_parse_dt, 'nargs': '*',
                    'default': 'start',
                    'help': 'output starting at FROM', 'dest': 'from'}),
        ('--to', {'action': 'store', 'type': custom_parse_dt, 'nargs': '*',
                  'default': 'end', 'help': 'output up to TO', 'dest': 'to'})
        ]

    timeunits = ['s', 'sec', 'm', 'min', 'h', 'hours', 'd', 'days', 'w',
                 'weeks', 'mo', 'months', 'y', 'years']
    weekdays = ['Mon', 'Tue', 'Wed', 'Thu', 'Fri', 'Sat', 'Sun']
    months = ['Jan', 'Feb', 'Mar', 'Apr', 'May', 'Jun', 'Jul', 'Aug', 'Sep',
              'Oct', 'Nov', 'Dec']

    dtRegexes = OrderedDict([
        # weekdays: see above
        ('weekday', r'|'.join(weekdays)),
        # month + day:  Jan 5, Oct 13, Sep 03, ...
        ('date', '(' + '|'.join(months) + ')' + r'\s+\d{1,2}'),
        ('word', r'now|start|end|today'),
        # 11:59, 1:13, 00:00, ...
        ('time2', r'\d{1,2}:\d{2,2}'),
        # 11:59:00, 1:13:12, 00:00:59, ...
        ('time3', r'\d{1,2}:\d{2,2}:\d{2,2}'),
        # offsets: +3min, -20s, +7days, ...
        ('offset', r'[\+-]\d+(' + '|'.join(timeunits) + ')'),
        ])

    def __init__(self, mlogfilter):
        BaseFilter.__init__(self, mlogfilter)
        self.fromReached = False
        self.toReached = False

        self.active = (('from' in self.mlogfilter.args and
                        self.mlogfilter.args['from'] != 'start') or
                       ('to' in self.mlogfilter.args and
                        self.mlogfilter.args['to'] != 'end'))

    def setup(self):
        """Get start end end date of logfile before starting to parse."""
        if self.mlogfilter.is_stdin:
            # assume this year (we have no other info)
            now = datetime.now()
            self.startDateTime = datetime(now.year, 1, 1, tzinfo=tzutc())
            self.endDateTime = datetime(MAXYEAR, 12, 31, tzinfo=tzutc())

        else:
            logfiles = self.mlogfilter.args['logfile']
            self.startDateTime = min([lf.start +
                                      timedelta(hours=self
                                                .mlogfilter
                                                .args['timezone'][i])
                                      for i, lf in enumerate(logfiles)])
            self.endDateTime = max([lf.end +
                                    timedelta(hours=self
                                              .mlogfilter.args['timezone'][i])
                                    for i, lf in enumerate(logfiles)])

        # now parse for further changes to from and to datetimes
        dtbound = DateTimeBoundaries(self.startDateTime, self.endDateTime)
        self.fromDateTime, self.toDateTime = dtbound(self.mlogfilter
                                                     .args['from'] or None,
                                                     self.mlogfilter
                                                     .args['to'] or None)

        # define start_limit for mlogfilter's fast_forward method
        self.start_limit = self.fromDateTime

        # for single logfile, get file seek position of `to` datetime
        if (len(self.mlogfilter.args['logfile']) == 1 and not
                self.mlogfilter.is_stdin):

            if self.mlogfilter.args['to'] != "end":
                # fast forward, get seek value, then reset file
                logfile = self.mlogfilter.args['logfile'][0]
                logfile.fast_forward(self.toDateTime)
                self.seek_to = logfile.filehandle.tell()
                logfile.filehandle.seek(0)
            else:
                self.seek_to = -1
        else:
            self.seek_to = False

    def accept(self, logevent):
        """
        Process line.

        Overwrite BaseFilter.accept() and return True if the provided
        logevent should be accepted (causing output), or False if not.
        """
        if self.fromReached and self.seek_to:
            if self.seek_to != -1:
                self.toReached = (self.mlogfilter.args['logfile'][0]
                                  .filehandle.tell() >= self.seek_to)
            return True
        else:
            # slow version has to check each datetime
            dt = logevent.datetime

            # if logevent has no datetime, accept if between --from and --to
            if dt is None:
                return self.fromReached

            if self.fromDateTime <= dt <= self.toDateTime:
                self.toReached = False
                self.fromReached = True
                return True

            elif dt > self.toDateTime:
                self.toReached = True
                return False

            else:
                return False

    def skipRemaining(self):
        """
        Skip remaining lines.

        Overwrite BaseFilter.skipRemaining() and return True if all lines
        from here to the end of the file should be rejected (no output).
        """
        return self.toReached



from functools import wraps

# tools without any external dependencies
from mtools.mlogfilter.mlogfilter import MLogFilterTool
from mtools.mloginfo.mloginfo import MLogInfoTool

tools = [MLogFilterTool, MLogInfoTool]


# mlaunch depends on pymongo
try:
    from mtools.mlaunch.mlaunch import MLaunchTool
    tools.append(MLaunchTool)
except ImportError:
    pass


# mplotqueries depends on matplotlib
try:
    from mtools.mplotqueries.mplotqueries import MPlotQueriesTool
    tools.append(MPlotQueriesTool)
except ImportError:
    pass


def all_tools(fn):
    """
    This is a decorator for test functions, that runs a loop over all command
    line tool classes imported above and passes each class to the test
    function.

    To use this decorator, the test function must accept a single
    parameter. Example:

        @all_tools
        def test_something(tool_cls):
            tool = tool_cls()
            # test tool here ...
    """
    @wraps(fn)
    # copies __name__ of the original function, pytest requires the name
    # to start with "test_"
    def new_func():
        for tool in tools:
            fn(tool)
    return new_func


def test_import_all():
    """
    Import all tools from mtools module.

    The tools that have external dependencies will only be imported if the
    dependencies are fulfilled. This test just passes by default because the
    imports are tested implicitly by loading this file.
    """
    pass



import os
from datetime import datetime
import re

from dateutil.tz import tzoffset, tzutc

import mtools
from mtools.util.logevent import LogEvent
from mtools.util.logfile import LogFile


class TestUtilLogFile(object):

    def setup(self):
        """Start up method for LogFile fixture."""

        # load logfile(s)
        self.logfile_path = os.path.join(os.path.dirname(mtools.__file__),
                                         'test/logfiles/', 'year_rollover.log')
        self.file_year_rollover = open(self.logfile_path, 'rb')
        self.current_year = datetime.now().year

    def test_len(self):
        """LogFile: test len() and iteration over LogFile method."""

        logfile = LogFile(self.file_year_rollover)
        length = len(logfile)

        i = 0
        for i, le in enumerate(logfile):
            assert isinstance(le, LogEvent)

        assert length == i + 1
        assert length == 1836

    def test_start_end(self):
        """LogFile: test .start and .end property work correctly."""

        logfile = LogFile(self.file_year_rollover)

        assert logfile.start == datetime(self.current_year - 1, 12, 30, 00,
                                         13, 1, 661000, tzutc())
        assert logfile.end == datetime(self.current_year, 1, 2, 23, 27, 11,
                                       720000, tzutc())

    def test_timezone(self):

        logfile_path = os.path.join(os.path.dirname(mtools.__file__),
                                    'test/logfiles/', 'mongod_26.log')
        mongod_26 = open(logfile_path, 'rb')

        logfile = LogFile(mongod_26)
        assert logfile.timezone == tzoffset(None, -14400)

    def test_rollover_detection(self):
        """LogFile: test datetime_format and year_rollover properties."""

        logfile = LogFile(self.file_year_rollover)
        assert logfile.datetime_format == "ctime"
        assert logfile.year_rollover == logfile.end

    def test_storage_engine_detection(self):
        """LogFile: test if the correct storage engine is detected."""

        logfile = LogFile(self.file_year_rollover)
        assert logfile.storage_engine is None

        logfile_path = os.path.join(os.path.dirname(mtools.__file__),
                                    'test/logfiles/', 'mongod_26.log')
        mmapv1 = open(logfile_path, 'rb')
        logfile = LogFile(mmapv1)
        assert logfile.storage_engine == 'mmapv1'

        # test for 3.0 WT detection
        logfile_path = os.path.join(os.path.dirname(mtools.__file__),
                                    'test/logfiles/', 'wiredtiger.log')
        wiredtiger = open(logfile_path, 'rb')
        logfile = LogFile(wiredtiger)
        assert logfile.storage_engine == 'wiredTiger'

        # test for 3.2 WT detection
        logfile_path = os.path.join(os.path.dirname(mtools.__file__),
                                    'test/logfiles/', 'mongod_328.log')
        wiredtiger = open(logfile_path, 'rb')
        logfile = LogFile(wiredtiger)
        assert logfile.storage_engine == 'wiredTiger'

    def test_rsinfo(self):
        """LogFile: test if replication info is detected (MongoDB 3.2+) """

        logfile_path = os.path.join(os.path.dirname(mtools.__file__),
                                    'test/logfiles/', 'rsinfo_36.log')
        rslog = open(logfile_path, 'rb')
        logfile = LogFile(rslog)
        assert logfile.repl_set == 'replset'
        assert logfile.repl_set_version == '1'
        assert logfile.repl_set_protocol == '1'

    def test_hostname_port(self):
        # mongod
        logfile_path = os.path.join(os.path.dirname(mtools.__file__),
                                    'test/logfiles/', 'mongod_26.log')
        mongod_26 = open(logfile_path, 'rb')

        logfile = LogFile(mongod_26)
        assert logfile.hostname == 'enter.local'
        assert logfile.port == '27019'

        # mongos
        logfile_path = os.path.join(os.path.dirname(mtools.__file__),
                                    'test/logfiles/', 'mongos.log')
        mongos = open(logfile_path, 'rb')

        logfile2 = LogFile(mongos)
        print(logfile2.hostname)
        assert logfile2.hostname == 'jimoleary.local'
        assert logfile2.port == '27017'

    def test_shard_info(self):
        """LogFile: test if sharding info is detected (MongoDB 3.6+) """

        # mongos log
        logfile_path = os.path.join(os.path.dirname(mtools.__file__),
                                    'test/logfiles/','sharding_360_mongos.log')
        mongos_log = open(logfile_path, 'rb')
        logfile = LogFile(mongos_log)
        shards = logfile.shards
        for name, repl_set in shards:
            assert re.match(r'shard\d+', name)
            assert re.match(r'localhost:270\d+', repl_set)

        # config log
        logfile_path = os.path.join(os.path.dirname(mtools.__file__),
                                    'test/logfiles/','sharding_360_CSRS.log')
        csrs_log = open(logfile_path, 'rb')
        logfile = LogFile(csrs_log)
        shards = logfile.shards
        for name, repl_set in shards:
            assert re.match(r'shard\d+', name)
            assert re.match(r'localhost:270\d+', repl_set)

        # shard log
        logfile_path = os.path.join(os.path.dirname(mtools.__file__),
                                    'test/logfiles/','sharding_360_shard.log')
        shard_log = open(logfile_path, 'rb')
        logfile = LogFile(shard_log)
        shards = logfile.shards
        for name, repl_set in shards:
            assert re.match(r'shard\d+', name)
            assert re.match(r'localhost:270\d+', repl_set)

    def test_shard_csrs(self):
        """LogFile: test if sharded cluster CSRS is detected (MongoDB 3.6+) """

        # mongos log
        logfile_path = os.path.join(os.path.dirname(mtools.__file__),
                                    'test/logfiles/','sharding_360_mongos.log')
        mongos_log = open(logfile_path, 'rb')
        logfile = LogFile(mongos_log)
        assert logfile.csrs == ('configRepl', 'localhost:27033')

        # config log
        logfile_path = os.path.join(os.path.dirname(mtools.__file__),
                                    'test/logfiles/','sharding_360_CSRS.log')
        csrs_log = open(logfile_path, 'rb')
        logfile = LogFile(csrs_log)
        assert logfile.csrs == ('configRepl', '[ { _id: 0, host: "localhost:27033",'
                                   ' arbiterOnly: false, buildIndexes: true, hidden: false,'
                                   ' priority: 1.0, tags: {}, slaveDelay: 0, votes: 1 } ]')

        # shard log
        logfile_path = os.path.join(os.path.dirname(mtools.__file__),
                                    'test/logfiles/','sharding_360_shard.log')
        shard_log = open(logfile_path, 'rb')
        logfile = LogFile(shard_log)
        assert logfile.csrs == ('configRepl', 'localhost:27033')

    def test_shard_chunk_migration_from(self):
        """
        LogFile: test if shard chunk migration (from) is detected (MongoDB 3.6+)
        """

        logfile_path = os.path.join(os.path.dirname(mtools.__file__),
                                    'test/logfiles/','sharding_360_shard.log')
        shard_log = open(logfile_path, 'rb')
        logfile = LogFile(shard_log)
        chunk_moved_from = logfile.chunks_moved_from[0]
        assert len(logfile.chunks_moved_from) == 2
        assert isinstance(chunk_moved_from[0], datetime)
        assert chunk_moved_from[1] == "min: { sku: MinKey }, max: { sku: 23153496 }"
        assert chunk_moved_from[2] == "shard02"
        assert chunk_moved_from[3] == "test.products"
        assert ('step 6 of 6', '57') in chunk_moved_from[4]
        assert chunk_moved_from[5] == "success"

    def test_shard_chunk_migration_to(self):
        """
        LogFile: test if shard chunk migration (to) is detected (MongoDB 3.6+)
        """

        logfile_path = os.path.join(os.path.dirname(mtools.__file__),
                                    'test/logfiles/','sharding_360_shard.log')
        shard_log = open(logfile_path, 'rb')
        logfile = LogFile(shard_log)
        chunk_moved_to = logfile.chunks_moved_to[0]

        assert len(logfile.chunks_moved_to) == 1
        assert isinstance(chunk_moved_to[0], datetime)
        assert chunk_moved_to[1] == "min: { sku: MinKey }, max: { sku: 23153496 }"
        assert chunk_moved_to[2] == "Unknown"
        assert chunk_moved_to[3] == "test.products"
        assert ('step 6 of 6', '213') in chunk_moved_to[4]
        assert chunk_moved_to[5] == "success"




from .scatter_type import ScatterPlotType, NScannedNPlotType
from .scatter_type import DocsExaminedPlotType, DurationLineType
from .event_type import EventPlotType, RSStatePlotType
from .range_type import RangePlotType
from .histogram_type import HistogramPlotType
from .connchurn_type import ConnectionChurnPlotType



from .clients_section import ClientSection
from .connection_section import ConnectionSection
from .cursors_section import CursorSection
from .distinct_section import DistinctSection
from .query_section import QuerySection
from .restart_section import RestartSection
from .rs_info_section import RsInfoSection
from .rs_state_section import RsStateSection
from .storagestats_section import StorageStatsSection
from .transactions_section import TransactionSection
from .sharding_section import ShardingSection


import re

from .base_filter import BaseFilter


class WordFilter(BaseFilter):
    """Accept only if line contains any of the words specified by --word."""

    filterArgs = [
        ('--word', {'action': 'store', 'nargs': '*',
                    'help': 'only output lines matching any of WORD'}),
        ]

    def __init__(self, mlogfilter):
        BaseFilter.__init__(self, mlogfilter)

        # extract all arguments passed into 'word'
        if 'word' in self.mlogfilter.args and self.mlogfilter.args['word']:
            self.words = self.mlogfilter.args['word'].split()
            self.active = True
        else:
            self.active = False

    def accept(self, logevent):
        """
        Process line.

        Overwrite BaseFilter.accept() and return True if the provided
        logevent should be accepted (causing output), or False if not.
        """
        for word in self.words:
            if re.search(word, logevent.line_str):
                return True
        return False



from collections import namedtuple
from operator import itemgetter

from .base_section import BaseSection
from mtools.util import OrderedDict
from mtools.util.grouping import Grouping
from mtools.util.print_table import print_table

LogTuple = namedtuple('LogTuple', ['namespace', 'operation', 'bytesRead',
                      'bytesWritten', 'timeReadingMicros', 'timeWritingMicros'])


def op_or_cmd(le):
    return le.operation if le.operation != 'command' else le.command


class StorageStatsSection(BaseSection):
    """StorageStatsSection class."""

    name = "Storage Statistics "

    def __init__(self, mloginfo):
        BaseSection.__init__(self, mloginfo)

        # add --storagestats flag to argparser
        helptext = 'outputs storage statistics for insert and update operations'
        self.mloginfo.argparser_sectiongroup.add_argument('--storagestats',
                                                          action='store_true',
                                                          help=helptext)

    @property
    def active(self):
        """Return boolean if this section is active."""
        return self.mloginfo.args['storagestats']

    def run(self):
        """Run this section and print out information."""
        grouping = Grouping(group_by=lambda x: (x.namespace, x.operation,
                                                x.bytesRead, x.bytesWritten,
                                                x.timeReadingMicros,
                                                x.timeWritingMicros))
        logfile = self.mloginfo.logfile

        if logfile.start and logfile.end:
            progress_start = self.mloginfo._datetime_to_epoch(logfile.start)
            progress_total = (self.mloginfo._datetime_to_epoch(logfile.end) -
                              progress_start)
        else:
            self.mloginfo.progress_bar_enabled = False

        for i, le in enumerate(logfile):
            # update progress bar every 1000 lines
            if self.mloginfo.progress_bar_enabled and (i % 1000 == 0):
                if le.datetime:
                    progress_curr = self.mloginfo._datetime_to_epoch(le.datetime)
                    if progress_total:
                        (self.mloginfo.update_progress(
                            float(progress_curr - progress_start) / progress_total))

            if (le.operation in ['update'] or le.command in ['insert']):
                lt = LogTuple(namespace=le.namespace, operation=op_or_cmd(le),
                              bytesRead=le.bytesRead, bytesWritten=le.bytesWritten,
                              timeReadingMicros=le.timeReadingMicros,
                              timeWritingMicros=le.timeWritingMicros)
                grouping.add(lt)

        grouping.sort_by_size()

        # clear progress bar again
        if self.mloginfo.progress_bar_enabled:
            self.mloginfo.update_progress(1.0)

        # no queries in the log file
        if not len(grouping):
            print('no statistics found.')
            return

        titles = ['namespace', 'operation', 'bytesRead', 'bytesWritten',
                  'timeReadingMicros', 'timeWritingMicros']
        table_rows = []

        for g in grouping:
            # calculate statistics for this group
            namespace, op, bytesRead, bytesWritten, timeReadingMicros, timeWritingMicros = g

            stats = OrderedDict()
            stats['namespace'] = namespace
            stats['operation'] = op
            stats['bytesRead'] = bytesRead
            stats['bytesWritten'] = bytesWritten
            stats['timeReadingMicros'] = timeReadingMicros
            stats['timeWritingMicros'] = timeWritingMicros

            table_rows.append(stats)

        print_table(table_rows, titles, uppercase_headers=False)
        print('')



#!/usr/bin/env python3

import pickle
import os
import re
import subprocess
import sys
from collections import defaultdict
from distutils.spawn import find_executable

from pymongo import MongoClient
from pymongo.errors import ConnectionFailure

from mtools.util.logcodeline import LogCodeLine

# Path to a local git clone from https://github.com/mongodb/mongo.git
mongodb_path = "/usr/local/src/mongo/"

git_path = find_executable('git')
pattern_id = 0


def source_files(mongodb_path):
    """Find source files."""
    for root, dirs, files in os.walk(mongodb_path):
        for filename in files:
            # skip files in dbtests folder
            if 'dbtests' in root:
                continue
            if filename.endswith(('.cpp', '.c', '.h')):
                yield os.path.join(root, filename)


def get_all_versions():
    pr = subprocess.Popen(git_path + " checkout master",
                          cwd=mongodb_path,
                          shell=True,
                          stdout=subprocess.PIPE,
                          stderr=subprocess.PIPE)
    pr.communicate()

    pr = subprocess.Popen(git_path + " tag",
                          cwd=mongodb_path,
                          shell=True,
                          stdout=subprocess.PIPE,
                          stderr=subprocess.PIPE)

    (out, error) = pr.communicate()
    versions = out.split()

    # only newer than 2.4.0
    versions = versions[versions.index("r2.4.0"):]

    # remove release candidates
    versions = [v for v in versions if "rc" not in v]

    # remove developer versions
    versions = [v for v in versions if re.search(r'\.[02468]\.', v)]

    # add master logs
    versions.append('master')

    return versions


def switch_version(version):
    print("Switching to %s" % version)
    pr = subprocess.Popen(git_path + " checkout %s" % version,
                          cwd=os.path.dirname(mongodb_path),
                          shell=True,
                          stdout=subprocess.PIPE,
                          stderr=subprocess.PIPE)

    (out, error) = pr.communicate()
    print(error)


def output_verbose(version, filename, lineno, line, statement,
                   matches, accepted, why):
    print("%10s %s %s:%s" % ("location:", version, filename, lineno))
    print("%10s %s" % ("line:", line))
    print("%10s %s" % ("statement:", statement))
    print("%10s %s" % ("matches:", matches))
    print("%10s %s" % ("accepted:", accepted))
    print("%10s %s" % ("reason:", why))
    print("----------------------------")
    print("\n")


def extract_logs(log_code_lines, current_version):
    global pattern_id
    log_templates = set()
    log_triggers = ([" log(", " LOG(", " LOGSOME", " warning()",
                     " error()", " out()", " problem()"])

    for filename in source_files(mongodb_path):
        f = open(filename, 'rb')

        # remove parent path
        filename = filename[len(mongodb_path):]

        lines = f.readlines()
        for lineno, line in enumerate(lines):
            trigger = next((t for t in log_triggers if t in line), None)

            if trigger:
                # extend line to wrap over line breaks until ; at end of line
                # is encountered
                statement = line
                current_lineno = lineno

                semicolon_match = None
                while not semicolon_match:
                    current_lineno += 1
                    if current_lineno >= len(lines):
                        break
                    statement += lines[current_lineno]
                    # match semicolon at end of line
                    # (potentially with whitespace between)
                    semicolon_match = re.search(';\s*$', statement,
                                                flags=re.MULTILINE)
                    if semicolon_match:
                        statement = statement[:semicolon_match.start()]
                        break

                # exclude triggers in comments (both // and /* */)
                trigger_pos = statement.find(trigger)
                newline_pos = statement.rfind("\n", 0, trigger_pos)
                if statement.find("//", newline_pos + 1, trigger_pos) != -1:
                    # output_verbose(current_version, filename, lineno,
                    # line, statement, "comment //")
                    continue
                comment_pos = statement.find("/*", 0, trigger_pos)
                if comment_pos != -1:
                    if statement.find("*/", comment_pos + 2,
                                      trigger_pos) == -1:
                        # output_verbose(current_version, filename, lineno,
                        # line, statement, "comment /* */")
                        continue

                statement = statement[statement.find(trigger) +
                                      len(trigger):].strip()

                # unescape strings
                # statement = statement.decode("string-escape")
                # print(statement)

                # remove compiler #ifdef .. #endif directives
                statement = re.sub(r'#ifdef.*?#endif', '',
                                   statement, flags=re.DOTALL)

                # filtering out conditional strings with
                # tertiary operator: ( ... ? ... : ... )
                statement = re.sub(r'\(.*?\?.*?\:.*?\)', '', statement)

                # split into stream tokens
                stream_tokens = statement.split("<<")

                # remove newlines from stream tokens
                stream_tokens = ([re.sub('\n', '', s).strip()
                                  for s in stream_tokens])

                matches = []
                for s in stream_tokens:
                    # match only non-empty strings with single / double quotes
                    match = re.match(r'"(.+?)"', s)
                    if match:
                        match = (re.sub(r'(\\t)|(\\n)|"', '',
                                        match.group(1)).strip())
                        matches.append(match)

                # get all double-quoted strings surrounded by << or ending in ;
                # print("s:::", statement)
                # matches = re.findall(r"<\s*\"(.*?)\"\s*(?:<|;)",
                #                      statement, flags=re.DOTALL)
                # print(matches)

                # remove tabs, double quotes and newlines and strip whitespace
                # from matches
                # matches = [re.sub(r'(\\t)|(\\n)|"', '', m).strip()
                #            for m in matches]
                # print(matches)

                # remove empty tokens
                matches = [m for m in matches if m]

                # skip empty matches
                if len(matches) == 0:
                    # output_verbose(current_version, filename, lineno, line,
                    #                statement, matches, False, "zero matches")
                    continue

                # skip matches with total character length < 3
                if len(''.join(matches)) < 3:
                    continue

                # skip matches consisting of single word, this will discard
                # some valid log lines but overall improves performance
                if len(matches) == 1 and " " not in matches[0]:
                    continue

                # special case causes trouble because of query operation lines
                if matches[0] == "query:":
                    # output_verbose(current_version, filename, lineno, line,
                    #                statement, matches, False,
                    #                "contains 'query:'")
                    continue

                loglevel = re.search(r'LOG\(\s*([0-9])\s*\)', line)
                if loglevel:
                    loglevel = int(loglevel.group(1))

                if trigger == ' log(':
                    loglevel = 0

                pattern = tuple(matches)

                # add to log_code_lines dict
                if pattern not in log_code_lines:
                    log_code_lines[pattern] = LogCodeLine(pattern, pattern_id)
                    pattern_id += 1

                # clean up values
                lineno = lineno + 1
                trigger = trigger.strip()

                log_code_lines[pattern].addMatch(current_version, filename,
                                                 lineno, loglevel, trigger)
                log_templates.add(pattern)

                # output_verbose(current_version, filename, lineno, line,
                #                statement, matches, True, "OK")

        f.close()

    return log_templates


if __name__ == '__main__':

    versions = get_all_versions()

    if len(sys.argv) > 1:
        versions = [sys.argv[1]]

    log_code_lines = {}
    logs_versions = defaultdict(list)
    print("parsing...")
    for v in versions:
        switch_version(v)
        logs = extract_logs(log_code_lines, v)
        print("version %s, %i lines extracted" % (v[1:], len(logs)))
        for l in logs:
            logs_versions[l].append(v)

    switch_version('master')

    # also store hashed by first word for quickly finding the related log lines
    logs_by_word = defaultdict(list)
    for lv in logs_versions:
        first_token = lv[0]
        split_words = first_token.split()
        logs_by_word[split_words[0]].append(lv)

    # now sort by number of tokens
    for lbw in logs_by_word:
        logs_by_word[lbw] = sorted(logs_by_word[lbw], key=lambda x: len(x),
                                   reverse=True)

    # for l in sorted(logs_versions):
    #     print(" <var> ".join(l), "found in:", ", ".join(logs_versions[l]))

    # write out to mongodb
    write_to_db = True
    try:
        mc = MongoClient()
        mc['log2code']['instances'].drop()
    except ConnectionFailure:
        write_to_db = False

    for pattern in log_code_lines:
        lcl = log_code_lines[pattern]

        first_token = pattern[0]
        split_words = first_token.split()

        instance = {
            '_id': lcl.pattern_id,
            'pattern': list(lcl.pattern),
            'first_word': split_words[0],
            'matches': []
        }

        for version in lcl.matches:
            matches = lcl.matches[version]
            for occ in matches:
                instance['matches'].append({
                    'version': version,
                    'file': occ[0],
                    'lineno': occ[1],
                    'loglevel': occ[2],
                    'trigger': occ[3]
                })

        if write_to_db:
            (mc['log2code']['instances']
             .update_one({'pattern': instance['pattern']}, instance,
                     upsert=True)())

    pickle.dump((versions, logs_versions, logs_by_word, log_code_lines),
                 open('log2code.pickle', 'wb'), -1)

    print("%i unique log messages imported and written to log2code.pickle"
          % len(log_code_lines))



from .base_section import BaseSection

try:
    from mtools.util.profile_collection import ProfileCollection
except ImportError:
    ProfileCollection = None


class RestartSection(BaseSection):
    """
    RestartSection class.

    This section determines if there were any restarts in the log file and
    prints out the times and version of the restarts found. It uses the
    information collected in LogFile so it doesn't have to walk the file
    manually.
    """

    name = "restarts"

    def __init__(self, mloginfo):
        BaseSection.__init__(self, mloginfo)

        # add --restarts flag to argparser
        helptext = 'outputs information about every detected restart'
        self.mloginfo.argparser_sectiongroup.add_argument('--restarts',
                                                          action='store_true',
                                                          help=helptext)

    @property
    def active(self):
        """Return boolean if this section is active."""
        return self.mloginfo.args['restarts']

    def run(self):
        """Run this section and print out information."""
        if ProfileCollection and isinstance(self.mloginfo.logfile,
                                            ProfileCollection):
            print("\n    not available for system.profile collections\n")
            return

        for version, logevent in self.mloginfo.logfile.restarts:
            print("   %s version %s"
                  % (logevent.datetime.strftime("%b %d %H:%M:%S"), version))

        if len(self.mloginfo.logfile.restarts) == 0:
            print("  no restarts found")



#!/usr/bin/env python3

def print_table(rows, override_headers=None, uppercase_headers=True):
    """All rows need to be a list of dictionaries, all with the same keys."""
    if len(rows) == 0:
        return
    keys = list(rows[0].keys())
    headers = override_headers or keys
    if uppercase_headers:
        rows = [dict(zip(keys,
                         map(lambda x: x.upper(), headers))), None] + rows
    else:
        rows = [dict(zip(keys, headers)), None] + rows

    lengths = [max(len(str(row[k]))
                   for row in rows if isinstance(row, dict)) for k in keys]
    tmp = ['{%s:%i}' % (h, l) for h, l in zip(keys[: -1], lengths[: -1])]
    tmp.append('{%s}' % keys[-1])
    template = (' ' * 4).join(tmp)

    for row in rows:
        if type(row) == str:
            print(row)
        elif row is None:
            print()
        elif isinstance(row, dict): 
            row = {k: v if v is not None else 'None' for k, v in row.items()}
            print(template.format(**row))
        else:
            print("Unhandled row type:", row)


if __name__ == '__main__':

    d = [{'a': '123', 'b': '654', 'c': 'foo'},
         {'a': '12ooo3', 'b': '654', 'c': 'foo'},
         {'a': '123', 'b': '65123124', 'c': 'foo'},
         {'a': '123', 'b': '654', 'c': 'fsadadsoo'},
         None,
         {'a': '123', 'b': '654', 'c': 'foo'}]

    print_table(d, ['long title here', 'foo', 'bar'])



import io
import json
import os
import re
import sys
from datetime import datetime, timedelta
from random import randrange

from dateutil import parser
import pytest

import mtools
from mtools.mlogfilter.mlogfilter import MLogFilterTool
from mtools.util.logevent import LogEvent
from mtools.util.logfile import LogFile


@pytest.fixture(scope="function", autouse=True)
def patch_io(monkeypatch, capsys):
    monkeypatch.setattr('sys.stdin', io.StringIO('my input'))
    sys.stdin.name = 'foo'
    sys.stdin.isatty = lambda: True


def random_date(start, end):
    """Return a random datetime between two datetime objects."""
    delta = end - start
    int_delta = (delta.days * 24 * 60 * 60) + delta.seconds
    random_second = randrange(int_delta)
    return start + timedelta(seconds=random_second)


class TestMLogFilter(object):
    """Test functionality around the mlogfilter tool."""

    def setup_method(self):
        """Start up method to create mlaunch tool and find free port."""
        self.tool = MLogFilterTool()

        self._test_base()

    def _test_base(self, filename='mongod_225.log'):
        # load logfile(s)
        self.logfile_path = os.path.join(os.path.dirname(mtools.__file__),
                                         'test/logfiles/', filename)
        self.logfile = LogFile(open(self.logfile_path, 'rb'))
        self.current_year = datetime.now().year

    def test_msToString(self):
        assert(self.tool._msToString(100) == '0hr 0min 0secs 100ms')
        assert(self.tool._msToString(1000) == '0hr 0min 1secs 0ms')
        assert(self.tool._msToString(100000) == '0hr 1min 40secs 0ms')
        assert(self.tool._msToString(10000000) == '2hr 46min 40secs 0ms')

    def test_from(self):
        random_start = random_date(self.logfile.start, self.logfile.end)
        self.tool.run('%s --from '
                      '%s' % (self.logfile_path,
                              random_start.strftime("%b %d %H:%M:%S.%f")))
        output = sys.stdout.getvalue()
        for line in output.splitlines():
            le = LogEvent(line)
            if not le.datetime:
                continue
            assert(le.datetime >= random_start)

    def test_from_iso8601_timestamp(self):
        random_start = random_date(self.logfile.start, self.logfile.end)
        self.tool.run('%s --from %s' % (self.logfile_path,
                                        random_start.isoformat()))
        output = sys.stdout.getvalue()
        for line in output.splitlines():
            le = LogEvent(line)
            if not le.datetime:
                continue
            assert(le.datetime >= random_start)

    def test_from_to(self):
        random_start = random_date(self.logfile.start, self.logfile.end)
        random_end = random_date(random_start, self.logfile.end)

        self.tool.run('%s --from %s --to '
                      '%s' % (self.logfile_path,
                              random_start.strftime("%b %d %H:%M:%S.%f"),
                              random_end.strftime("%b %d %H:%M:%S.%f")))
        output = sys.stdout.getvalue()
        for line in output.splitlines():
            le = LogEvent(line)
            if not le.datetime:
                continue
            assert(le.datetime >= random_start and le.datetime <= random_end)

    def test_from_to_26_log(self):
        logfile_26_path = os.path.join(os.path.dirname(mtools.__file__),
                                       'test/logfiles/', 'mongod_26.log')
        logfile_26 = LogFile(open(logfile_26_path, 'rb'))

        random_start = random_date(logfile_26.start, logfile_26.end)
        random_end = random_date(random_start + timedelta(minutes=1),
                                 logfile_26.end + timedelta(minutes=1))

        print('%s %s' % (random_start, random_end))
        print('%s %s' % (logfile_26.start, logfile_26.end))

        self.tool.run('%s --from %s --to '
                      '%s' % (logfile_26_path,
                              random_start.strftime("%b %d %H:%M:%S.%f"),
                              random_end.strftime("%b %d %H:%M:%S.%f")))
        output = sys.stdout.getvalue()
        assert len(output.splitlines()) > 0

        at_least_one = False
        for line in output.splitlines():
            le = LogEvent(line)
            if not le.datetime:
                continue
            at_least_one = True
            assert(le.datetime >= random_start and le.datetime <= random_end)
        assert at_least_one

    def test_from_to_stdin(self):

        year = datetime.now().year
        start = datetime(year, 8, 5, 20, 45)
        end = datetime(year, 8, 5, 21, 1)
        self.tool.is_stdin = True
        self.tool.run('%s --from %s --to '
                      '%s' % (self.logfile_path,
                              start.strftime("%b %d %H:%M:%S.%f"),
                              end.strftime("%b %d %H:%M:%S.%f")))
        self.tool.is_stdin = False

        output = sys.stdout.getvalue()
        for line in output.splitlines():
            le = LogEvent(line)
            assert(le.datetime >= start and le.datetime <= end)

    def test_json(self):
        """Output with --json is in JSON format."""
        self.tool.run('%s --json' % self.logfile_path)
        output = sys.stdout.getvalue()
        for line in output.splitlines():
            line_dict = json.loads(line)
            assert(line_dict)
            assert(type(line_dict) == dict)

    def test_shorten_50(self):
        self.tool.run('%s --shorten 50' % self.logfile_path)
        output = sys.stdout.getvalue()
        for line in output.splitlines():
            assert(len(line) <= 50)

    def test_shorten_default(self):
        self.tool.run('%s --shorten' % self.logfile_path)
        output = sys.stdout.getvalue()
        for line in output.splitlines():
            assert(len(line) <= 200)

    def test_merge_same(self):
        file_length = len(self.logfile)
        self.tool.run('%s %s' % (self.logfile_path, self.logfile_path))
        output = sys.stdout.getvalue()
        lines = output.splitlines()
        assert len(lines) == 2 * file_length
        for prev, next in zip(lines[:-1], lines[1:]):
            prev_le = LogEvent(prev)
            next_le = LogEvent(next)
            if not prev_le.datetime or not next_le.datetime:
                continue
            assert prev_le.datetime <= next_le.datetime

    def test_merge_markers(self):
        file_length = len(self.logfile)
        self.tool.run('%s %s --markers foo bar' % (self.logfile_path,
                                                   self.logfile_path))
        output = sys.stdout.getvalue()
        lines = output.splitlines()
        assert len([l for l in lines if l.startswith('foo')]) == file_length
        assert len([l for l in lines if l.startswith('bar')]) == file_length

    def test_merge_invalid_markers(self):
        try:
            self.tool.run('%s %s --markers foo bar baz' % (self.logfile_path,
                                                           self.logfile_path))
        except SystemExit as e:
            assert 'Number of markers not the same' in str(e)

    def test_exclude(self):
        file_length = len(self.logfile)
        tool = MLogFilterTool()
        tool.run('%s --slow 300' % self.logfile_path)

        tool = MLogFilterTool()
        tool.run('%s --slow 300 --exclude' % self.logfile_path)
        output = sys.stdout.getvalue()
        lines_total = len(output.splitlines())

        assert lines_total == file_length

    def test_end_reached(self):
        self.tool.run('%s --from Jan 3015 --to +10min' % self.logfile_path)
        output = sys.stdout.getvalue()
        assert output.strip() == ''

    def test_human(self):
        self.tool.run('%s --slow --thread conn8 --human' % self.logfile_path)
        output = sys.stdout.getvalue().rstrip()
        assert(output.endswith('(0hr 0min 1secs 324ms) 1,324ms'))
        assert('cursorid:7,776,022,515,301,717,602' in output)

    def test_slow_fast(self):
        self.tool.run('%s --slow 145 --fast 500' % self.logfile_path)
        output = sys.stdout.getvalue()
        assert(len(output.splitlines()) > 0)
        for line in output.splitlines():
            le = LogEvent(line)
            assert(le.duration >= 145 and le.duration <= 500)

    @pytest.mark.xfail(raises=SystemExit)
    def test_invalid_log(self):
        # load text file
        invalid_logfile_path = os.path.join(os.path.dirname(mtools.__file__),
                                            '../', 'requirements.txt')
        self.tool.run('%s' % invalid_logfile_path)

    def test_scan(self):
        # load tablescan logfile
        scn_logfile_path = os.path.join(os.path.dirname(mtools.__file__),
                                        'test/logfiles/', 'collscans.log')

        self.tool.run('%s --scan' % scn_logfile_path)
        output = sys.stdout.getvalue()
        lines = output.splitlines()
        assert len(lines) == 3

    def test_accept_nodate(self):
        self.tool.is_stdin = True
        self.tool.run('%s --from Aug 5 %d 20:53:50 --to '
                      '+5min' % (self.logfile_path, self.current_year - 1))
        self.tool.is_stdin = False

        output = sys.stdout.getvalue()
        lines = output.splitlines()
        assert any('=== a line without a datetime ===' in l for l in lines)

    def _test_thread(self, path, thread):
        self.tool.run('%s --thread mongosMain' % path)
        output = sys.stdout.getvalue()
        for line in output.splitlines():
            le = LogEvent(line)
            assert(le.thread == thread)
            md = re.match("^.* connection accepted from [0-9\.:]+ "
                          "#(?P<conn>[0-9]+) ", le.line_str)
            if md is None:
                assert(le.conn is None)
            else:
                assert(le.conn == "conn" + md.group('conn'))

    def _test_thread_conn1(self, path, thread):
        self.tool.run('%s --thread conn1' % path)
        output = sys.stdout.getvalue()
        for line in output.splitlines():
            le = LogEvent(line)
            md = re.match("^.* connection accepted from [0-9\.:]+ "
                          "#(?P<conn>[0-9]+) ", le.line_str)
            assert(le.conn == 'conn1')
            if md is None:
                assert(le.thread == 'conn1')
            else:
                assert(le.thread == thread)

    def test_thread(self):
        self._test_thread(self.logfile_path, 'initandlisten')

    def test_thread_conn1(self):
        self._test_thread_conn1(self.logfile_path, 'initandlisten')

    def test_thread_mongos(self):
        mongos_path = os.path.join(os.path.dirname(mtools.__file__),
                                   'test/logfiles/', 'mongos.log')
        self._test_thread(mongos_path, 'mongosMain')

    def test_thread_mongos_conn1(self):
        mongos_path = os.path.join(os.path.dirname(mtools.__file__),
                                   'test/logfiles/', 'mongos.log')
        self._test_thread_conn1(mongos_path, 'mongosMain')

    def test_no_timestamp_format(self):
        self.tool.run('%s --timestamp-format none --timezone 5'
                      % self.logfile_path)
        output = sys.stdout.getvalue()
        for line in output.splitlines():
            le = LogEvent(line)
            if le.datetime:
                assert le.datetime_format == 'ctime-pre2.4'

    def test_operation(self):
        self.tool.run('%s --operation insert' % self.logfile_path)
        output = sys.stdout.getvalue()
        for line in output.splitlines():
            le = LogEvent(line)
            assert(le.operation == 'insert')

    def test_multiple_operations(self):
        self.tool.run('%s --operation insert query' % self.logfile_path)
        output = sys.stdout.getvalue()
        for line in output.splitlines():
            le = LogEvent(line)
            assert(le.operation in ['insert', 'query'])

    def test_invalid_timezone_args(self):
        try:
            self.tool.run('%s --timezone 1 2 3' % self.logfile_path)
        except SystemExit as e:
            assert "Invalid number of timezone parameters" in str(e)

    def test_verbose(self):
        self.tool.run('%s --slow --verbose' % self.logfile_path)
        output = sys.stdout.getvalue()
        lines = output.splitlines()
        assert lines[0].startswith('command line arguments')
        assert any(line.startswith('active filters: SlowFilter')
                   for line in lines)

    def test_namespace(self):
        self.tool.run('%s --namespace local.oplog.rs' % self.logfile_path)
        output = sys.stdout.getvalue()
        for line in output.splitlines():
            le = LogEvent(line)
            assert(le.namespace == 'local.oplog.rs')

    def test_pattern(self):
        # test that pattern is correctly parsed, reordered and compared to
        # logevent pattern
        self.tool.run('%s --pattern {ns:1,_id:1,host:1}' % self.logfile_path)
        output = sys.stdout.getvalue()
        for line in output.splitlines():
            le = LogEvent(line)
            assert(le.pattern == '{"_id": 1, "host": 1, "ns": 1}')

    def test_command(self):
        self.tool.run('%s --command dropDatabase deleteIndexes'
                      % self.logfile_path)
        output = sys.stdout.getvalue()
        for line in output.splitlines():
            le = LogEvent(line)
            assert(le.command in ['dropDatabase', 'deleteIndexes'])

    def test_planSummary(self):
        logfile_path = os.path.join(os.path.dirname(mtools.__file__),
                                    'test/logfiles/', 'mongod_26_corrupt.log')
        self.tool.run('%s --planSummary IXSCAN' % logfile_path)
        output = sys.stdout.getvalue()
        lines = output.splitlines()
        assert(len(lines) > 0)
        for line in lines:
            le = LogEvent(line)
            assert(le.planSummary == "IXSCAN")

    def test_word(self):
        self.tool.run('%s --word lock' % self.logfile_path)
        output = sys.stdout.getvalue()
        for line in output.splitlines():
            assert('lock' in line)

    def test_mask_end(self):
        mask_path = os.path.join(os.path.dirname(mtools.__file__),
                                 'test/logfiles/', 'mask_centers.log')

        event1 = parser.parse("Mon Aug  5 20:27:15 UTC")
        event2 = parser.parse("Mon Aug  5 20:30:09 UTC")
        mask_size = randrange(10, 60)
        padding = timedelta(seconds=mask_size / 2)

        self.tool.run('%s --mask %s --mask-size %i' % (self.logfile_path,
                                                       mask_path, mask_size))
        output = sys.stdout.getvalue()
        for line in output.splitlines():
            le = LogEvent(line)
            assert((le.datetime >= event1 - padding and
                    le.datetime <= event1 + padding) or
                   (le.datetime >= event2 - padding and
                    le.datetime <= event2 + padding))

    def test_mask_start(self):
        mask_path = os.path.join(os.path.dirname(mtools.__file__),
                                 'test/logfiles/', 'mask_centers.log')

        event1 = parser.parse("Mon Aug  5 20:27:15 UTC")
        duration1 = timedelta(seconds=75)
        event2 = parser.parse("Mon Aug  5 20:30:09 UTC")
        mask_size = randrange(10, 60)
        padding = timedelta(seconds=mask_size / 2)

        self.tool.run('%s --mask %s --mask-size %i --mask-center start'
                      % (self.logfile_path, mask_path, mask_size))
        output = sys.stdout.getvalue()
        for line in output.splitlines():
            le = LogEvent(line)
            assert((le.datetime >= event1 - duration1 - padding and
                    le.datetime <= event1 - duration1 + padding) or
                   (le.datetime >= event2 - padding and
                    le.datetime <= event2 + padding))

    def test_mask_both(self):
        mask_path = os.path.join(os.path.dirname(mtools.__file__),
                                 'test/logfiles/', 'mask_centers.log')

        event1 = parser.parse("Mon Aug  5 20:27:15 UTC")
        duration1 = timedelta(seconds=75)
        event2 = parser.parse("Mon Aug  5 20:30:09 UTC")
        mask_size = randrange(10, 60)
        padding = timedelta(seconds=mask_size / 2)

        self.tool.run('%s --mask %s --mask-size %i --mask-center both'
                      % (self.logfile_path, mask_path, mask_size))
        output = sys.stdout.getvalue()
        for line in output.splitlines():
            le = LogEvent(line)
            assert((le.datetime >= event1 - duration1 - padding and
                    le.datetime <= event1 + padding) or
                   (le.datetime >= event2 - padding and
                    le.datetime <= event2 + padding))

    @pytest.mark.xfail(raises=SystemExit)
    def test_no_logfile(self):
        """Test that not providing at least 1 log file throws clean error."""

        self.tool.run('--from Jan 1')

    def test_year_rollover_1(self):
        """
        Test that mlogfilter works correctly with year-rollovers in logfiles
        with ctime (1).
        """

        # load year rollover logfile
        yro_logfile_path = os.path.join(os.path.dirname(mtools.__file__),
                                        'test/logfiles/', 'year_rollover.log')

        self.tool.run('%s --from Jan 1 %d --timestamp-format iso8601-utc'
                      % (yro_logfile_path, self.current_year))
        output = sys.stdout.getvalue()
        for line in output.splitlines():
            assert line.startswith('%d-' % self.current_year)

    def test_year_rollover_2(self):
        """
        Test that mlogfilter works correctly with year-rollovers in logfiles
        with ctime (2).
        """

        # load year rollover logfile
        yro_logfile_path = os.path.join(os.path.dirname(mtools.__file__),
                                        'test/logfiles/', 'year_rollover.log')
        self.tool.run('%s --from Dec 31 --to +1day --timestamp-format '
                      'iso8601-utc' % yro_logfile_path)
        output = sys.stdout.getvalue()
        assert len(output.splitlines()) > 0
        for line in output.splitlines():
            assert line.startswith('%d-' % (self.current_year - 1))

    def test_level_225(self):
        """Test that mlogfilter works levels on older logs."""

        self.tool.run('%s --level D ' % self.logfile_path)
        output = sys.stdout.getvalue()
        assert len(output.splitlines()) == 0

    def test_component_225(self):
        """Test that mlogfilter works components on older logs."""

        self.tool.run('%s --component ACCESS ' % self.logfile_path)
        output = sys.stdout.getvalue()
        assert len(output.splitlines()) == 0

    @pytest.mark.xfail(raises=SystemExit)
    def test_invalid_level(self):
        self._test_base('mongod_278.log')
        self.tool.run(' %s --level C' % self.logfile_path)

    @pytest.mark.xfail(raises=SystemExit)
    def test_invalid_component(self):
        self._test_base('mongod_278.log')
        self.tool.run(' %s --component FAKE' % self.logfile_path)

    def _test_levels(self, level, expected=1):
        self._test_base('mongod_278.log')
        self.tool.run(' %s --level %s' % (self.logfile_path, level))
        output = sys.stdout.getvalue()
        assert len(output.splitlines()) == expected

    def _test_components(self, component, expected=1):
        self._test_base('mongod_278.log')
        self.tool.run(' %s --component %s' % (self.logfile_path, component))
        output = sys.stdout.getvalue()
        assert len(output.splitlines()) == expected


def _add_component_test(cls, name, component, expected):
    """Meta program new tests."""
    def test_method(self):
        self._test_components(component, expected)
    test_method.__doc__ = name
    test_method.__name__ = name
    setattr(cls, test_method.__name__, test_method)


_add_component_test(TestMLogFilter, 'test_total_component', 'TOTAL', 1)
_add_component_test(TestMLogFilter, 'test_s2_component', 'GEO', 1)
_add_component_test(TestMLogFilter, 'test_all_component',
                    " ".join(LogEvent.log_components), 25)
_add_component_test(TestMLogFilter, 'test_dash_component', '-', 10)
_add_component_test(TestMLogFilter, 'test_access_component', 'ACCESS', 1)
_add_component_test(TestMLogFilter, 'test_commands_component', 'COMMAND', 1)
_add_component_test(TestMLogFilter, 'test_indexing_component', 'INDEX', 1)
_add_component_test(TestMLogFilter, 'test_network_component', 'NETWORK', 1)
_add_component_test(TestMLogFilter, 'test_query_component', 'QUERY', 1)
_add_component_test(TestMLogFilter, 'test_replsets_component', 'REPL', 1)
_add_component_test(TestMLogFilter, 'test_sharding_component', 'SHARDING', 1)
_add_component_test(TestMLogFilter, 'test_storage_component', 'STORAGE', 4)
_add_component_test(TestMLogFilter, 'test_journal_component', 'JOURNAL', 1)
_add_component_test(TestMLogFilter, 'test_writes_component', 'WRITE', 1)


def _add_level_test(cls, name, level, expected=1):
    """
    meta program new tests
    """
    def test_method(self):
        self._test_levels(level, expected)
    test_method.__doc__ = name
    test_method.__name__ = name
    setattr(cls, test_method.__name__, test_method)


_add_level_test(TestMLogFilter, 'test_all_levels',
                " ".join(LogEvent.log_levels), 25)
_add_level_test(TestMLogFilter, 'test_D_levels', 'D')
_add_level_test(TestMLogFilter, 'test_F_levels', 'F')
_add_level_test(TestMLogFilter, 'test_E_levels', 'E')
_add_level_test(TestMLogFilter, 'test_W_levels', 'W')
_add_level_test(TestMLogFilter, 'test_I_levels', 'I', 20)
_add_level_test(TestMLogFilter, 'test_U_levels', 'U')

# output = sys.stdout.getvalue().strip()



#!/usr/bin/env python3

import datetime
import inspect
import sys

import mtools.mloginfo.sections as sections
from mtools.util.cmdlinetool import LogFileTool


class MLogInfoTool(LogFileTool):

    def __init__(self):
        """Constructor: add description to argparser."""
        LogFileTool.__init__(self, multiple_logfiles=True, stdin_allowed=False)

        self.argparser.description = ('Extracts general information from '
                                      'logfile and prints it to stdout.')
        self.argparser.add_argument('--debug', action='store_true',
                                    help=('show debug output '
                                          '(depends on info section)'))
        self.argparser.add_argument('--verbose', action='store_true',
                                    help=('show more verbose output '
                                          '(depends on info section)'))

        inf = 'info sections'
        cmds = ('Below commands activate additional info sections for the '
                'log file.')
        self.argparser_sectiongroup = self.argparser.add_argument_group(inf,
                                                                        cmds)

        # add all filter classes from the filters module
        self.sections = ([c[1](self)
                          for c in inspect.getmembers(sections,
                                                      inspect.isclass)])

    def run(self, arguments=None):
        """Print useful information about the log file."""
        LogFileTool.run(self, arguments)

        if (self.args['logfile'] is None or len(self.args['logfile']) == 0):
            self.argparser.print_usage()
            print("\nERROR: At least one logfile argument must be provided")
            self.argparser.exit()

        for i, self.logfile in enumerate(self.args['logfile']):
            if i > 0:
                print("\n ------------------------------------------\n")

            if self.logfile.datetime_format == 'ctime-pre2.4':
                # no milliseconds when datetime format doesn't support it
                start_time = (self.logfile.start.strftime("%Y %b %d %H:%M:%S")
                              if self.logfile.start else "unknown")
                end_time = (self.logfile.end.strftime("%Y %b %d %H:%M:%S")
                            if self.logfile.start else "unknown")
            else:
                # include milliseconds
                start_time = (self.logfile.start.strftime("%Y %b %d "
                                                          "%H:%M:%S.%f")[:-3]
                              if self.logfile.start else "unknown")
                end_time = (self.logfile.end.strftime("%Y %b %d "
                                                      "%H:%M:%S.%f")[:-3]
                            if self.logfile.start else "unknown")

            print("     source: %s" % self.logfile.name)
            print("       host: %s"
                  % (self.logfile.hostname + ':' + str(self.logfile.port)
                     if self.logfile.hostname else "unknown"))
            print("      start: %s" % (start_time))
            print("        end: %s" % (end_time))

            print("date format: %s" % self.logfile.datetime_format)

            # self.logfile.timezone is a dateutil.tzinfo object
            tzdt = datetime.datetime.now(self.logfile.timezone)
            if (tzdt.tzname()):
                timezone = tzdt.tzname()
            else:
                timezone = f"UTC {tzdt.strftime('%z')}"
            print("   timezone: %s" % timezone)
            print("     length: %s" % len(self.logfile))
            print("     binary: %s" % (self.logfile.binary or "unknown"))

            version = (' -> '.join(self.logfile.versions) or "unknown")

            # if version is unknown, go by date
            if version == 'unknown':
                if self.logfile.datetime_format == 'ctime-pre2.4':
                    version = '< 2.4 (no milliseconds)'
                elif self.logfile.datetime_format == 'ctime':
                    version = '>= 2.4.x ctime (milliseconds present)'
                elif (self.logfile.datetime_format == "iso8601-utc" or
                      self.logfile.datetime_format == "iso8601-local"):
                    if self.logfile.has_level:
                        version = '>= 3.0 (iso8601 format, level, component)'
                    else:
                        version = '= 2.6.x (iso8601 format)'

            print("    version: %s" % version)
            print("    storage: %s"
                  % (self.logfile.storage_engine or 'unknown'))

            # now run all sections
            for section in self.sections:
                if section.active:
                    print("\n%s" % section.name.upper())
                    section.run()


def main():
    tool = MLogInfoTool()
    tool.run()
    return 0  # we need to return an integer


if __name__ == '__main__':
    sys.exit(main())



import sys
from datetime import datetime, timedelta
from random import randrange

from dateutil import parser
import pytest

import os

import mtools
from mtools.mplotqueries.mplotqueries import MPlotQueriesTool
from mtools.util.logevent import LogEvent
from mtools.util.logfile import LogFile


@pytest.fixture(scope="function", autouse=True)
def patch_io(monkeypatch, capsys):
    monkeypatch.setattr('sys.stdin', io.StringIO('my input'))
    sys.stdin.name = 'foo'
    sys.stdin.isatty = lambda: True

    
class TestMPlotQueries(object):

    def setup_method(self):
        """Startup method to create mplotqueries tool."""
        self.tool = MPlotQueriesTool()
        self._test_init()

    def _test_init(self, filename='mongod_225.log'):
        # load logfile(s)
        self.logfile_path = os.path.join(os.path.dirname(mtools.__file__),'test/logfiles/', filename)
        self.logfile = LogFile(open(self.logfile_path, 'rb'))

    @pytest.mark.skip(reason='Skipping interactive test')
    def test_dns(self, filename='mongod_4.0.10_slowdns.log'):
        self.logfile_path = os.path.join(os.path.dirname(mtools.__file__),'test/logfiles/', filename)
        self.tool.run('%s --dns' % self.logfile_path)
        output = sys.stdout.getvalue()
        lines = output.splitlines()
        assert any(map(lambda line: 'SCATTER plot' in line, lines))

    @pytest.mark.skip(reason='Skipping interactive test')
    def test_checkpoints(self, filename='mongod_4.0.10_slowcheckpoints.log'):
        self.logfile_path = os.path.join(os.path.dirname(mtools.__file__), 'test/logfiles/', filename)
        self.tool.run('%s --checkpoints' % self.logfile_path)
        output = sys.stdout.getvalue()
        lines = output.splitlines()
        assert any(map(lambda line: 'SCATTER plot' in line, lines))

    @pytest.mark.skip(reason='Skipping interactive test')
    def test_oplog(self, filename='mongod_4.0.10_slowoplogs.log'):
        self.logfile_path = os.path.join(os.path.dirname(mtools.__file__), 'test/logfiles/', filename)
        self.tool.run('%s --oplog --group operation' % self.logfile_path)
        output = sys.stdout.getvalue()
        lines = output.splitlines()
        assert any('SCATTER plot' in line for line in lines)

    @pytest.mark.skip(reason='Skipping interactive test')
    def test_storagestats(self, filename='mongod_4.0.10_storagestats.log'):
        self.logfile_path = os.path.join(os.path.dirname(mtools.__file__),'test/logfiles/', filename)
        self.tool.run('%s --storagestats --yaxis bytesRead' % self.logfile_path)
        output = sys.stdout.getvalue()
        lines = output.splitlines()
        assert any(map(lambda line: 'SCATTER plot' in line, lines))


from mtools.util.log2code import Log2CodeConverter

logline1 = ("Thu Nov 14 17:58:43.898 [rsStart] replSet info Couldn't load "
            "config yet. Sleeping 20sec and will try again.")
logline2 = ("Thu Nov 14 17:58:43.917 [initandlisten] connection accepted "
            "from 10.10.0.38:37233 #10 (4 connections now open)")

l2cc = Log2CodeConverter()


def test_log2code():
    fixed, variable = l2cc(logline1)
    assert fixed
    assert fixed.matches["r2.4.9"] == [('src/mongo/db/repl/rs.cpp', 790, 0,
                                        'log(')]



import re
from collections import defaultdict

from .base_section import BaseSection

try:
    from mtools.util.profile_collection import ProfileCollection
except ImportError:
    ProfileCollection = None


class ConnectionSection(BaseSection):
    """
    ConnectionSection class.

    This section goes through the logfile and extracts information about
    opened and closed connections.
    """

    name = "connections"

    def __init__(self, mloginfo):
        BaseSection.__init__(self, mloginfo)

        helptext = 'outputs information about opened and closed connections'
        self.mloginfo.argparser_sectiongroup.add_argument('--connections',
                                                          action='store_true',
                                                          help=helptext)
        helptext = ('outputs helpful statistics for connection '
                    'duration (min/max/avg)')
        self.mloginfo.argparser_sectiongroup.add_argument('--connstats',
                                                          action='store_true',
                                                          help=helptext)

    @property
    def active(self):
        """Return boolean if this section is active."""
        return(self.mloginfo.args['connections'] or
               self.mloginfo.args['connstats'])

    def run(self):
        """Run this section and print out information."""
        if ProfileCollection and isinstance(self.mloginfo.logfile,
                                            ProfileCollection):
            print("\n    not available for system.profile collections\n")
            return

        ip_opened = defaultdict(lambda: 0)
        ip_closed = defaultdict(lambda: 0)

        socket_exceptions = 0

        START_TIME_EMPTY = -11
        END_TIME_ALREADY_FOUND = -111
        MIN_DURATION_EMPTY = 9999999999
        MAX_DURATION_EMPTY = -1

        end_connid_pattern = re.compile(r'\[conn(\d+)\]')

        genstats = self.mloginfo.args['connstats']
        if genstats:
            connections_start = defaultdict(lambda: START_TIME_EMPTY)
            ipwise_sum_durations = defaultdict(lambda: 0)
            ipwise_count = defaultdict(lambda: 0)
            ipwise_min_connection_duration = defaultdict(lambda:
                                                         MIN_DURATION_EMPTY)
            ipwise_max_connection_duration = defaultdict(lambda:
                                                         MAX_DURATION_EMPTY)

            min_connection_duration = MIN_DURATION_EMPTY
            max_connection_duration = MAX_DURATION_EMPTY

            sum_durations = 0
            fullconn_counts = 0

        for logevent in self.mloginfo.logfile:
            line = logevent.line_str

            pos = line.find('connection accepted')
            if pos != -1:
                # connection was opened, increase counter
                tokens = line[pos:pos + 100].split(' ')
                if tokens[3] == 'anonymous':
                    ip = 'anonymous'
                else:
                    ip, _ = tokens[3].split(':')
                ip_opened[ip] += 1

                if genstats:
                    connid = tokens[4].strip('#')
                    dt = logevent.datetime

                    # Sanity checks
                    if connid.isdigit() is False or dt is None:
                        continue

                    if connections_start[connid] != START_TIME_EMPTY:
                        errmsg = ("Multiple start datetimes found for the "
                                  "same connection ID. Consider analysing one "
                                  "log sequence.")
                        raise NotImplementedError(errmsg)

                    connections_start[connid] = dt

            pos = line.find('end connection')
            if pos != -1:
                # connection was closed, increase counter
                tokens = line[pos:pos + 100].split(' ')
                if tokens[2] == 'anonymous':
                    ip = 'anonymous'
                else:
                    ip, _ = tokens[2].split(':')
                ip_closed[ip] += 1

                if genstats:

                    # Sanity check
                    if end_connid_pattern.search(line, re.M | re.I) is None:
                        continue

                    # The connection id value is stored just before end
                    # connection -> [conn385] end connection
                    end_connid = (end_connid_pattern.
                                  search(line, re.M | re.I).group(1))
                    dt = logevent.datetime

                    # Sanity checks
                    if (end_connid.isdigit() is False or dt is None or
                            connections_start[end_connid] == START_TIME_EMPTY):
                        continue

                    if connections_start[end_connid] == END_TIME_ALREADY_FOUND:
                        errmsg = ("Multiple end datetimes found for the same "
                                  "connection ID %s. Consider analysing one "
                                  "log sequence.")
                        raise NotImplementedError(errmsg % (end_connid))

                    dur = dt - connections_start[end_connid]
                    dur_in_sec = dur.seconds

                    if dur_in_sec < min_connection_duration:
                        min_connection_duration = dur_in_sec

                    if dur_in_sec > max_connection_duration:
                        max_connection_duration = dur_in_sec

                    if dur_in_sec < ipwise_min_connection_duration[ip]:
                        ipwise_min_connection_duration[ip] = dur_in_sec

                    if dur_in_sec > ipwise_max_connection_duration[ip]:
                        ipwise_max_connection_duration[ip] = dur_in_sec

                    sum_durations += dur.seconds
                    fullconn_counts += 1

                    ipwise_sum_durations[ip] += dur_in_sec
                    ipwise_count[ip] += 1

                    connections_start[end_connid] = END_TIME_ALREADY_FOUND

            if "SocketException" in line:
                socket_exceptions += 1

        # calculate totals
        total_opened = sum(ip_opened.values())
        total_closed = sum(ip_closed.values())

        unique_ips = set(ip_opened.keys())
        unique_ips.update(ip_closed.keys())

        # output statistics
        print("     total opened: %s" % total_opened)
        print("     total closed: %s" % total_closed)
        print("    no unique IPs: %s" % len(unique_ips))
        print("socket exceptions: %s" % socket_exceptions)
        if genstats:
            if fullconn_counts > 0:
                print("overall average connection duration(s): %s"
                      % (sum_durations / fullconn_counts))
                print("overall minimum connection duration(s): %s"
                      % min_connection_duration)
                print("overall maximum connection duration(s): %s"
                      % max_connection_duration)
            else:
                print("overall average connection duration(s): -")
                print("overall minimum connection duration(s): -")
                print("overall maximum connection duration(s): -")
        print('')

        for ip in sorted(unique_ips, key=lambda x: ip_opened[x], reverse=True):
            opened = ip_opened[ip] if ip in ip_opened else 0
            closed = ip_closed[ip] if ip in ip_closed else 0

            if genstats:
                covered_count = (
                    ipwise_count[ip]
                    if ip in ipwise_count
                    else 1)
                connection_duration_ip = (
                    ipwise_sum_durations[ip]
                    if ip in ipwise_sum_durations
                    else 0)
                ipwise_min_connection_duration_final = (
                    ipwise_min_connection_duration[ip]
                    if ipwise_min_connection_duration[ip] != MIN_DURATION_EMPTY
                    else 0)
                ipwise_max_connection_duration_final = (
                    ipwise_max_connection_duration[ip]
                    if ipwise_max_connection_duration[ip] != MAX_DURATION_EMPTY
                    else 0)

                print("%-15s  opened: %-8i  closed: %-8i dur-avg(s): %-8i "
                      "dur-min(s): %-8i dur-max(s): %-8i"
                      % (ip, opened, closed,
                         connection_duration_ip / covered_count,
                         ipwise_min_connection_duration_final,
                         ipwise_max_connection_duration_final))
            else:
                print("%-15s  opened: %-8i  closed: %-8i"
                      % (ip, opened, closed))

        print('')



from .base_section import BaseSection


class RsInfoSection(BaseSection):
    """
    RsInfoSection class.

    This section determines if there is any Replica Set infomation like the
    replset name in the log file and prints the available information.
    """

    name = "rsinfo"

    def __init__(self, mloginfo):
        BaseSection.__init__(self, mloginfo)

        # add --rsinfo flag to argparser
        helptext = 'outputs replica set config information'
        self.mloginfo.argparser_sectiongroup.add_argument('--rsinfo',
                                                          action='store_true',
                                                          help=helptext)

    @property
    def active(self):
        """Return boolean if this section is active."""
        return self.mloginfo.args['rsinfo']

    def run(self):
        """Run this section and print out information."""
        if self.mloginfo.logfile.repl_set:
            print("    rs name: %s" % self.mloginfo.logfile.repl_set)
            print(" rs members: %s"
                  % (self.mloginfo.logfile.repl_set_members
                     if self.mloginfo.logfile.repl_set_members
                     else "unknown"))
            print(" rs version: %s"
                  % (self.mloginfo.logfile.repl_set_version
                     if self.mloginfo.logfile.repl_set_version
                     else "unknown"))
            print("rs protocol: %s"
                  % (self.mloginfo.logfile.repl_set_protocol
                     if self.mloginfo.logfile.repl_set_protocol
                     else "unknown"))
        else:
            print("  no rs info changes found")



from .base_filter import BaseFilter


class TableScanFilter(BaseFilter):
    """
    TableScanFilter class.

    Accept only if the line contains a nscanned:[0-9] nreturned:[0-9] where
    the ratio of nscanned:nreturned is > 100 and nscanned > 10000.
    """

    filterArgs = [
        ('--scan', {
            'action': 'store_true',
            'help': ('only output lines which have poor index usage '
                     '(nscanned>10000 and ratio of nscanned to '
                     'nreturned>100); to see confirmed collection scans, '
                     'use --planSummary COLLSCAN.')
            })
        ]

    def __init__(self, mlogfilter):
        BaseFilter.__init__(self, mlogfilter)

        if 'scan' in self.mlogfilter.args:
            self.active = self.mlogfilter.args['scan']

    def accept(self, logevent):
        """
        Process line.

        Overwrite BaseFilter.accept() and return True if the provided
        logevent should be accepted (causing output), or False if not.
        """
        ns = logevent.nscanned
        nr = logevent.nreturned

        if ns is not None and nr is not None:
            if nr == 0:
                # avoid division by 0 errors
                nr = 1
            return (ns > 10000 and ns / nr > 100)

        return False



#!/usr/bin/env python3
"""Input source utility."""


class InputSource(object):
    """Object definition for InputSource."""

    def __iter__(self):
        """Iterate over log events."""
        pass

    def fast_forward(self, dt):
        """Fast forward over log events."""
        pass



from .datetime_filter import DateTimeFilter
from .fast_filter import FastFilter
from .slow_filter import SlowFilter
from .logline_filter import LogLineFilter
from .tablescan_filter import TableScanFilter
from .word_filter import WordFilter
from .mask_filter import MaskFilter
from .transaction_filter import TransactionFilter



#!/usr/bin/env python3
"""Command line tool utility."""

import argparse
import codecs
import datetime
import os
import re
import signal
import sys

from dateutil.tz import tzutc

from mtools.util.logfile import LogFile
from mtools.version import __version__

try:
    from mtools.util.profile_collection import ProfileCollection

    class InputSourceAction(argparse.FileType):
        """
        Extend the FileType class from the argparse module.

        Try to open the file and pass the handle to a new LogFile object, but
        if that's not possible it will catch the exception and interpret the
        string as a MongoDB URI and try to connect to the database. In that
        case, it will return a ProfileCollection object.

        Both derive from the same base class InputSource and support iteration
        over LogEvents.
        """

        def __call__(self, string):
            """Open log file or MongoDB database."""

            try:
                # catch filetype and return LogFile object
                filehandle = argparse.FileType.__call__(self, string)
                return LogFile(filehandle)

            except argparse.ArgumentTypeError:
                # not a file, try open as MongoDB database
                m = re.match('^(\w+)(?::(\d+))?(?:/([a-zA-Z0-9._-]+))?$',
                             string)
                if m:
                    hostname, port, namespace = m.groups()
                    port = int(port) if port else 27017
                    namespace = namespace or 'test.system.profile'
                    if '.' in namespace:
                        database, collection = namespace.split('.', 1)
                    else:
                        database = namespace
                        collection = 'system.profile'

                    if (hostname == 'localhost' or
                            re.match('\d+\.\d+\.\d+\.\d+', hostname)):
                        return ProfileCollection(hostname, port, database,
                                                 collection)

                raise argparse.ArgumentTypeError("can't open %s as file or "
                                                 "MongoDB connection string."
                                                 % string)

except ImportError:
    class InputSourceAction(argparse.FileType):
        """Extend the FileType class from the argparse module."""

        def __call__(self, string):
            """Open log file."""
            try:
                # catch filetype and return LogFile object
                filehandle = argparse.FileType.__call__(self, string)
                return LogFile(filehandle)
            except argparse.ArgumentTypeError:
                raise argparse.ArgumentTypeError("can't open %s" % string)


class BaseCmdLineTool(object):
    """
    Base class for any mtools command line tool.

    Adds --version flag and basic control flow.
    """

    def __init__(self):
        """
        Constructor.

        Any inheriting class should add a description to the argparser and
        extend it with additional arguments as needed.
        """
        # define argument parser and add version argument
        self.argparser = argparse.ArgumentParser()
        self.argparser.add_argument(
            '--version', 
            action='version',
            version=f'''mtools version {__version__} || Python {sys.version}'''
        )
        self.argparser.add_argument('--no-progressbar', action='store_true',
                                    default=False,
                                    help='disables progress bar')
        self.is_stdin = not sys.stdin.isatty()

    def run(self, arguments=None, get_unknowns=False):
        """
        Init point to execute the script.

        If `arguments` string is given, will evaluate the arguments, else
        evaluates sys.argv. Any inheriting class should extend the run method
        (but first calling BaseCmdLineTool.run(self)).
        """

        # redirect PIPE signal to quiet kill script, if not on Windows
        if os.name != 'nt':
            signal.signal(signal.SIGPIPE, signal.SIG_DFL)
        if get_unknowns:
            if arguments:
                self.args, self.unknown_args = (self.argparser.parse_known_args
                                                (args=arguments.split()))
            else:
                (self.args,
                 self.unknown_args) = self.argparser.parse_known_args()
            self.args = vars(self.args)
        else:
            if arguments:
                myargs = arguments.split()
                self.args = vars(self.argparser.parse_args
                                 (args=myargs))
            else:
                self.args = vars(self.argparser.parse_args())

        self.progress_bar_enabled = (not (self.args['no_progressbar'] or
                                          self.is_stdin))

    def _datetime_to_epoch(self, dt):
        """Convert the datetime to unix epoch (properly)."""
        if dt:
            td = (dt - datetime.datetime.fromtimestamp(0, tzutc()))
            # don't use total_seconds(), that's only available in 2.7
            total_secs = int((td.microseconds +
                              (td.seconds + td.days * 24 * 3600) *
                              10**6) / 10**6)
            return total_secs
        else:
            return 0

    def update_progress(self, progress, prefix=''):
        """
        Print a progress bar for longer-running scripts.

        The progress value is a value between 0.0 and 1.0. If a prefix is
        present, it will be printed before the progress bar.
        """
        total_length = 40

        if progress == 1.:
            sys.stderr.write('\r' + ' ' * (total_length + len(prefix) + 50))
            sys.stderr.write('\n')
            sys.stderr.flush()
        else:
            bar_length = int(round(total_length * progress))
            sys.stderr.write('\r%s [%s%s] %.1f %% '
                             % (prefix, '=' * bar_length,
                                ' ' * (total_length - bar_length),
                                progress * 100))
            sys.stderr.flush()


class LogFileTool(BaseCmdLineTool):
    """Base class for any mtools tool that acts on logfile(s)."""

    def __init__(self, multiple_logfiles=False, stdin_allowed=True):
        """Add logfile(s) and stdin option to the argument parser."""
        BaseCmdLineTool.__init__(self)

        self.multiple_logfiles = multiple_logfiles
        self.stdin_allowed = stdin_allowed

        arg_opts = {'action': 'store', 'type': InputSourceAction('rb')}

        if self.multiple_logfiles:
            arg_opts['nargs'] = '*'
            arg_opts['help'] = 'logfile(s) to parse'
        else:
            arg_opts['help'] = 'logfile to parse'

        if self.is_stdin:
            if not self.stdin_allowed:
                raise SystemExit("this tool can't parse input from stdin.")

            arg_opts['const'] = LogFile(sys.stdin)
            arg_opts['action'] = 'store_const'
            if 'type' in arg_opts:
                del arg_opts['type']
            if 'nargs' in arg_opts:
                del arg_opts['nargs']
        self.argparser.add_argument('logfile', **arg_opts)


if __name__ == '__main__':
    tool = LogFileTool(multiple_logfiles=True, stdin_allowed=True)
    tool.run()
    print(tool.args)
    # for line in tool.args['logfile']:
    #     print(line)



import argparse

import numpy as np

from mtools.mplotqueries.plottypes.base_type import BasePlotType
from mtools.util.log2code import Log2CodeConverter

try:
    from matplotlib.dates import date2num, num2date
except ImportError as error:
    raise ImportError("Can't import matplotlib. See "
                      "https://matplotlib.org/users/installing.html "
                      "for instructions on how to install matplotlib."
                      "Error: " + str(error))


class HistogramPlotType(BasePlotType):
    """
    Plot a histogram plot over all logevents.

    The bucket size can be specified with the --bucketsize or -b parameter.
    Unit is in seconds.
    """

    plot_type_str = 'histogram'
    timeunits = {'sec': 1, 's': 1, 'min': 60, 'm': 1, 'hour': 3600, 'h': 3600,
                 'day': 86400, 'd': 86400}
    sort_order = 1
    default_group_by = 'namespace'
    l2cc = Log2CodeConverter()

    def __init__(self, args=None, unknown_args=None):
        BasePlotType.__init__(self, args, unknown_args)

        # parse arguments further to get --bucketsize argument
        argparser = argparse.ArgumentParser("mplotqueries --type histogram")
        argparser.add_argument('--bucketsize', '-b', action='store',
                               metavar='SIZE',
                               help="histogram bucket size in seconds",
                               default=60)
        argparser.add_argument('--no-stacked', action='store_true',
                               help=("switch graph mode from stacked "
                                     "histogram (default) to side-by-side "
                                     "histograms."), default=False)
        sub_args = vars(argparser.parse_args(unknown_args))

        self.logscale = args['logscale']
        # get bucket size, either as int (seconds) or as string
        # (see timeunits above)
        bs = sub_args['bucketsize']
        try:
            self.bucketsize = int(bs)
        except ValueError:
            self.bucketsize = self.timeunits[bs]
        self.barstacked = not sub_args['no_stacked']

        self.ylabel = "# lines per %i second bin" % self.bucketsize

    def accept_line(self, logevent):
        """
        Return True for each line.

        We bucket everything. Filtering has to be done before passing to this
        type of plot.
        """
        return True

    def log2code(self, logevent):
        codeline = self.l2cc(logevent.line_str)
        if codeline:
            return ' ... '.join(codeline.pattern)
        else:
            return None

    def plot_group(self, group, idx, axis):
        raise NotImplementedError("Not implemented for histogram plots.")

    def plot(self, axis, ith_plot, total_plots, limits):
        """
        Plot the histogram as a whole over all groups.

        Do not plot as individual groups like other plot types.
        """
        print(self.plot_type_str.upper() + " plot")
        print("%5s %9s  %s" % ("id", " #points", "group"))

        for idx, group in enumerate(self.groups):
            print("%5s %9s  %s" % (idx + 1, len(self.groups[group]), group))

        print('')

        datasets = []
        colors = []
        minx = np.inf
        maxx = -np.inf

        for idx, group in enumerate(self.groups):
            x = date2num([logevent.datetime
                          for logevent in self.groups[group]])
            minx = min(minx, min(x))
            maxx = max(maxx, max(x))
            datasets.append(x)
            color, marker = self.color_map(group)
            colors.append(color)

        if total_plots > 1:
            # if more than one plot, move histogram to twin axis on the right
            twin_axis = axis.twinx()
            twin_axis.set_ylabel(self.ylabel)
            axis.set_zorder(twin_axis.get_zorder() + 1)  # put ax ahead of ax2
            axis.patch.set_visible(False)  # hide the 'canvas'
            axis = twin_axis

        n_bins = max(1, int((maxx - minx) * 24. * 60. * 60. / self.bucketsize))
        if n_bins > 1000:
            # warning for too many buckets
            print("warning: %i buckets, will take a while to render. "
                  "consider increasing --bucketsize." % n_bins)

        n, bins, artists = axis.hist(datasets, bins=n_bins, align='mid',
                                     log=self.logscale,
                                     histtype="barstacked"
                                     if self.barstacked else "bar",
                                     color=colors, edgecolor="none",
                                     linewidth=0, alpha=0.8, picker=True,
                                     label=map(str, self.groups.keys()))

        # scale current y-axis to match min and max values
        axis.set_ylim(np.min(n), np.max(n))

        # add meta-data for picking
        if len(self.groups) > 1:
            for g, group in enumerate(self.groups.keys()):
                for i in range(len(artists[g])):
                    artists[g][i]._mt_plot_type = self
                    artists[g][i]._mt_group = group
                    artists[g][i]._mt_n = n[g][i]
                    if self.barstacked:
                        artists[g][i]._mt_n -= (n[g - 1][i] if g > 0 else 0)

                    artists[g][i]._mt_bin = bins[i]
        else:
            for i in range(len(artists)):
                artists[i]._mt_plot_type = self
                artists[i]._mt_group = group
                artists[i]._mt_n = n[i]
                artists[i]._mt_bin = bins[i]

        return artists

    def clicked(self, event):
        """Print group name and number of items in bin."""
        group = event.artist._mt_group
        n = event.artist._mt_n
        dt = num2date(event.artist._mt_bin)
        print("%4i %s events in %s sec beginning at %s"
              % (n, group, self.bucketsize, dt.strftime("%b %d %H:%M:%S")))



from .base_filter import BaseFilter
from mtools.util.logevent import LogEvent
from mtools.util.pattern import json2pattern


def custom_parse_array(value):
    return list(set(value.split()))


class LogLineFilter(BaseFilter):
    """LogLineFilter class."""

    filterArgs = [
        ('--component', {
            'nargs': '*',
            'action': 'store',
            'type': str.upper,
            'choices': LogEvent.log_components,
            'metavar': 'CM',
            'help': ('only output log lines with component CM '
                     '(multiple values are allowed).')
        }),
        ('--level', {
            'nargs': '*',
            'action': 'store',
            'metavar': 'LL',
            'choices': LogEvent.log_levels,
            'help': ('only output log lines with loglevel LL '
                     '(multiple values are allowed).')
        }),
        ('--namespace', {
            'nargs': '*',
            'action': 'store',
            'metavar': 'NS',
            'help': ('only output log lines on namespace NS '
                     '(multiple values are allowed).')
        }),
        ('--operation', {
            'nargs': '*',
            'action': 'store',
            'metavar': 'OP',
            'choices': LogEvent.log_operations,
            'help': ('only output log lines of type OP '
                     '(multiple values are allowed).')
        }),
        ('--thread', {
            'nargs': '*',
            'action': 'store',
            'help': ('only output log lines of thread THREAD '
                     '(multiple values are allowed).')
        }),
        ('--pattern', {
            'action': 'store',
            'help': ('only output log lines with a query pattern PATTERN '
                     '(only applies to queries, getmores, updates, removes).')
        }),
        ('--command', {
            'nargs': '*',
            'action': 'store',
            'help': ('only output log lines which are commands of the given '
                     'type. Examples: "distinct", "isMaster", '
                     '"replSetGetStatus" (multiple values are allowed).')
        }),
        ('--planSummary', {
            'nargs': '*',
            'action': 'store',
            'type': str.upper,
            'metavar': 'PS',
            'help': ('only output log lines which match the given plan '
                     'summary values (multiple values are allowed).')
            })
        ]

    def __init__(self, mlogfilter):
        BaseFilter.__init__(self, mlogfilter)

        self.components = None
        self.levels = None
        self.namespaces = None
        self.operations = None
        self.threads = None
        self.commands = None
        self.pattern = None
        self.planSummaries = None

        if ('component' in self.mlogfilter.args and
                self.mlogfilter.args['component']):
            self.components = custom_parse_array(self.mlogfilter
                                                 .args['component'])
            self.active = True
        if 'level' in self.mlogfilter.args and self.mlogfilter.args['level']:
            self.levels = custom_parse_array(self.mlogfilter.args['level'])
            self.active = True
        if ('namespace' in self.mlogfilter.args and
                self.mlogfilter.args['namespace']):
            self.namespaces = custom_parse_array(self.mlogfilter
                                                 .args['namespace'])
            self.active = True
        if ('operation' in self.mlogfilter.args and
                self.mlogfilter.args['operation']):
            self.operations = custom_parse_array(self.mlogfilter
                                                 .args['operation'])
            self.active = True
        if ('command' in self.mlogfilter.args and
                self.mlogfilter.args['command']):
            self.commands = custom_parse_array(self.mlogfilter.args['command'])
            self.active = True
        if 'thread' in self.mlogfilter.args and self.mlogfilter.args['thread']:
            self.threads = custom_parse_array(self.mlogfilter.args['thread'])
            self.active = True
        if ('pattern' in self.mlogfilter.args and
                self.mlogfilter.args['pattern']):
            self.pattern = json2pattern(self.mlogfilter.args['pattern'])
            self.active = True
            if self.pattern is None:
                raise SystemExit("ERROR: cannot parse pattern \"%s\" as a JSON"
                                 " string" % self.mlogfilter.args['pattern'])
        if ('planSummary' in self.mlogfilter.args and
                self.mlogfilter.args['planSummary']):
            self.planSummaries = custom_parse_array(self.mlogfilter
                                                    .args['planSummary'])
            self.active = True

    def accept(self, logevent):
        """
        Process line.

        Overwrite BaseFilter.accept() and return True if the provided
        logevent should be accepted (causing output), or False if not.
        """
        # if several filters are active, all have to agree
        if self.components and logevent.component not in self.components:
            return False
        if self.levels and logevent.level not in self.levels:
            return False
        if self.namespaces and logevent.namespace not in self.namespaces:
            return False
        if self.commands and logevent.command not in self.commands:
            return False
        if self.operations and logevent.operation not in self.operations:
            return False
        if self.threads:
            if (logevent.thread not in self.threads and
                    logevent.conn not in self.threads):
                return False
        if self.pattern and logevent.pattern != self.pattern:
            return False
        if (self.planSummaries and
                logevent.planSummary not in self.planSummaries):
            return False

        return True



#!/usr/bin/env python

import argparse
import os
import re
import sys
import bson
import wiredtiger

from mtools.util.cmdlinetool import BaseCmdLineTool
from mtools.version import __version__

codec_options = bson.codec_options.CodecOptions(uuid_representation=bson.binary.STANDARD)


class MTransferTool(BaseCmdLineTool):
    def __init__(self):
        BaseCmdLineTool.__init__(self)

        self.argparser.description = ('Import and export databases between MongoDB deployments '
                                      'for WiredTiger storage with directoryPerDB configuration.')

        self.argparser.add_argument('--dbpath', dest='dbpath', default='.', nargs=1,
                                    help='MongoDB database path')

        self.argparser.add_argument('--force', action='store_true',
                                    help='ignore safety checks')

        self.argparser.add_argument('--verbose', action='store_true',
                                    help='enable verbose output')

        self.argparser.add_argument('command', choices=['export', 'import'])

        self.argparser.add_argument('database', nargs=1, type=str,
                                    help='name of the database to export / import')

    def run(self, arguments=None):
        BaseCmdLineTool.run(self, arguments)

        self.dbpath = self.args['dbpath'][0]
        self.force = self.args['force']
        self.verbose = self.args['verbose']

        # Read storage.bson, sanity check.
        try:
            storage_raw = open(os.path.join(self.dbpath, 'storage.bson'), 'rb').read()
        except Exception as e:
            sys.stderr.write(f'''Failed to open storage.bson in "{self.dbpath}": {e}\n''')
            return

        settings = bson.decode(storage_raw)["storage"]["options"]
        if not settings["directoryPerDB"]:
            sys.stderr.write('Requires a database created with --directoryperdb\n')
            return
        if settings["directoryForIndexes"] or settings.get("groupCollections", False):
            sys.stderr.write('Incompatible storage settings detected: '
                             'directoryForIndexes or groupCollections\n')
            if not self.force:
                return

        self.database = self.args['database'][0]
        self.nsprefix = self.database + '.'

        mtransfer_dir = os.path.join(self.dbpath, self.database)
        mtransfer_file = os.path.join(mtransfer_dir, 'mtransfer.bson')

        if self.args['command'] == 'export':
            if not os.path.exists(mtransfer_dir):
                msg = f'''Expected source directory "{mtransfer_dir}" does not exist. '''
                msg = msg + 'Check the database name is correct.\n'
                sys.stderr.write(msg)
                return
            if not self.force and os.path.exists(mtransfer_file):
                sys.stderr.write(f'''Output file "{mtransfer_file}" already exists\n''')
                return
            with open(mtransfer_file, 'wb') as outf:
                self.doExport(outf)
        elif self.args['command'] == 'import':
            if not os.path.exists(mtransfer_dir):
                msg = f'''Expected target directory "{mtransfer_dir}" does not exist. '''
                msg = msg + 'Check the database name is correct.\n'
                sys.stderr.write(msg)
                return
            if not os.path.exists(mtransfer_file):
                sys.stderr.write(f'''Cannot import: mtransfer file "{mtransfer_file}" does not exist.\n''')
                return
            with open(mtransfer_file, 'rb') as inf:
                self.doImport(inf)

    def message(self, msg):
        if self.verbose:
            print(msg)

    def doExport(self, outf):
        # Attempt to connect to the specified WiredTiger database
        try:
            conn = wiredtiger.wiredtiger_open(
                self.dbpath,
                'log=(compressor=snappy,path=journal,recover=error),readonly=true')
        except Exception as e:
            sys.stderr.write(f'''Failed to open dbpath "{self.dbpath}": {e}\n''')
            return

        session = conn.open_session()

        # Find all collections in the database
        catalog = session.open_cursor('table:_mdb_catalog')
        sizeStorer = session.open_cursor('table:sizeStorer')
        wtMeta = session.open_cursor('metadata:')
        wtMetaCreate = session.open_cursor('metadata:create')
        for _, meta_raw in catalog:
            meta = bson.decode(meta_raw, codec_options=codec_options)
            ns = meta[u'ns']
            if not ns or not ns.startswith(self.nsprefix):
                continue
            assert ns == meta[u'md'][u'ns']

            # Iterate through indexes first
            indexes = {}
            for idxName, idxIdent in meta[u'idxIdent'].items():
                ident = str('table:' + idxIdent)
                filename = ident[len('table:'):] + '.wt'
                file_ident = 'file:' + filename
                wtmeta_file = wtMeta[file_ident]
                wtmeta_table = wtMetaCreate[ident]
                basename = filename[len(self.nsprefix):]
                indexes[idxName] = {'filename': basename,
                                    'wtmeta_table': wtmeta_table,
                                    'wtmeta_file': wtmeta_file}

            collname = ns[len(self.nsprefix):]
            ident = str('table:' + meta[u'ident'])
            size = bson.decode(sizeStorer[ident.encode()])
            filename = ident[len('table:'):] + '.wt'
            file_ident = 'file:' + filename
            wtmeta_file = wtMeta[file_ident]
            wtmeta_table = wtMetaCreate[ident]
            basename = filename[len(self.nsprefix):]
            export = {
                'collname': collname,
                'filename': basename,
                'mdb_catalog': meta,
                'sizeStorer': size,
                'wtmeta_table': wtmeta_table,
                'wtmeta_file': wtmeta_file,
                'indexes': indexes,
                'version': __version__,
            }
            self.message(str(export))
            outf.write(bson.encode(export, codec_options=codec_options))

        conn.close()

    def doImport(self, inf):
        try:
            conn = wiredtiger.wiredtiger_open(
                self.dbpath,
                'log=(enabled=false,compressor=snappy,path=journal,recover=error)')
        except Exception as e:
            sys.stderr.write(f'''Failed to open dbpath "{self.dbpath}": {e}\n''')
            return

        try:
            self._doImport(conn, inf)
        except Exception as e:
            sys.stderr.write(f'''Import failed: {e}''')

        print('Import complete')

        conn.close()

    def _doImport(self, conn, inf):
        app_metadata_re = re.compile(r'app_metadata=\(.*?\)')

        session = conn.open_session()
        session.begin_transaction()

        catalog = session.open_cursor('table:_mdb_catalog')
        sizeStorer = session.open_cursor('table:sizeStorer')
        wtMeta = session.open_cursor('metadata:', None, 'readonly=false')

        # Get the maximum file ID in the WT catalog: we will be appending
        session.create('file:_mtransfer')
        newfile_meta = wtMeta['file:_mtransfer']
        self.message(f'''Got new file metadata "{newfile_meta}"''')
        session.drop('file:_mtransfer')
        file_id = int(re.search(r',id=(\d+),', newfile_meta).group(1))

        # Get the maximum ID in the MDB catalog: we will be appending
        catalog.prev()
        maxID = catalog.get_key()

        for export in bson.decode_file_iter(inf, codec_options=codec_options):
            if not os.path.exists(
                    os.path.join(self.dbpath, self.database, export['filename'])):
                sys.stderr.write(
                    f'''File "{export['filename']}" referenced in export missing during import'''
                )
                if not self.force:
                    return

            if not self.force and export['version'] != __version__:
                sys.stderr.write(
                    f'''Database was exported with mtools version {export['version']}, '''
                    f'''current version {__version__} may not be compatible'''
                )
                return

            # Figure out the new namespace
            ns = self.database + '.' + export['collname']

            # First process the indexes
            idxIdent = {}
            for idxName, idx in export['indexes'].items():
                ident = self.database + '/' + idx['filename'][:-3]
                table_uri = 'table:' + ident
                colgroup_uri = 'colgroup:' + ident
                file_uri = 'file:' + ident + '.wt'
                # Do a regular "session.create" for the table, then overwrite
                # the "file:" metadata with the original
                app_metadata = app_metadata_re.search(idx['wtmeta_file']).group(0)
                # For older style index metadata, update the namespace
                app_metadata = re.sub(r'"ns" : ".*?"', f'''"ns" : "{ns}"''', app_metadata)
                self.message(f'''For index "{idxName}", app_metadata = "{app_metadata}"''')
                wtMeta[table_uri] = (app_metadata +
                                     ',colgroups=,collator=,columns=,key_format=u,value_format=u')
                wtMeta[colgroup_uri] = (app_metadata +
                                        ',collator=,columns=,source="' + file_uri + '",type=file')
                wtMeta[file_uri] = (idx['wtmeta_file'] + ',' + app_metadata +
                                    (f''',id={file_id:d}'''))
                file_id += 1
                idxIdent[idxName] = ident
                self.message(f'''Adding index "{idxName}" with ident "{ident}"''')

            # Figure out the WT URIs
            ident = self.database + '/' + export['filename'][:-3]
            table_uri = 'table:' + ident
            colgroup_uri = 'colgroup:' + ident
            file_uri = 'file:' + ident + '.wt'

            # Do a regular "session.create" for the table, then overwrite the
            # "file:" metadata with the original
            app_metadata = app_metadata_re.search(export['wtmeta_file']).group(0)
            self.message(f'''For collection "{ns}", app_metadata = "{app_metadata}"''')
            wtMeta[table_uri] = (app_metadata +
                                 ',colgroups=,collator=,columns=,key_format=q,value_format=u')
            wtMeta[colgroup_uri] = (app_metadata +
                                    ',collator=,columns=,source="' + file_uri + '",type=file')
            wtMeta[file_uri] = (export['wtmeta_file'] + ',' + app_metadata +
                                (f''',id={file_id:d}'''))
            file_id += 1

            sizeStorer[ident.encode()] = bson.encode(export['sizeStorer'])

            # Fix the catalog entry to refer to the new namespace and new table names
            catalog_entry = export['mdb_catalog']
            catalog_entry[u'ns'] = ns
            catalog_entry[u'md'][u'ns'] = ns
            catalog_entry[u'ident'] = ident
            catalog_entry[u'idxIdent'] = idxIdent
            for i in range(len(catalog_entry[u'md'][u'indexes'])):
                catalog_entry[u'md'][u'indexes'][i][u'spec'][u'ns'] = ns
            maxID += 1
            self.message(f'''Adding catalog entry {maxID} -> {catalog_entry}''')
            catalog[maxID] = bson.encode(catalog_entry, codec_options=codec_options)

        session.commit_transaction()


def main():
    tool = MTransferTool()
    tool.run()
    return 0  # we need to return an integer


if __name__ == '__main__':
    sys.exit(main())



from datetime import timedelta

from .datetime_filter import DateTimeFilter
from mtools.util.cmdlinetool import InputSourceAction


class MaskFilter(DateTimeFilter):
    """
    MaskFilter class.

    This filter takes an argument `--mask <LOGFILE>` and another optional
    argument `--mask-size <SECS>`. It will read <LOGFILE> and for each of the
    lines extract the datetimes (let's call these "events"). It will add some
    padding for each of these events, <SECS>/2 seconds on either side.
    MaskFilter will then accept every line from the original log file
    (different to <LOGFILE>), that lies within one of these masked intervals.

    This feature is very useful to find all correlating lines to certain
    events.

    For example, find all assertions in a log file, then find all log lines
    surrounding these assertions:

        grep "assert" mongod.log > assertions.log
        mlogfilter mongod.log --mask assertions.log --mask-size 60
    """

    filterArgs = [
        ('--mask', {'action': 'store', 'type': InputSourceAction('rb'),
                    'help': ('source (log file or system.profile db) '
                             'to create the filter mask.')}),
        ('--mask-size', {'action': 'store', 'type': int, 'default': 60,
                         'help': ('mask size in seconds around each filter '
                                  'point (default: 60 secs, 30 on each side '
                                  'of the event)')}),
        ('--mask-center', {'action': 'store',
                           'choices': ['start', 'end', 'both'],
                           'default': 'end',
                           'help': ('mask center point for events with '
                                    'duration (default: end). If both is '
                                    'chosen, all events from start to end '
                                    'are returned.')})
        ]

    def __init__(self, mlogfilter):
        """
        Constructor.

        Init superclass and mark this filter active if `mask`
        argument is present.
        """
        DateTimeFilter.__init__(self, mlogfilter)
        self.active = ('mask' in self.mlogfilter.args and
                       self.mlogfilter.args['mask'] is not None)
        if self.active:
            self.mask_end_reached = False
            self.mask_source = self.mlogfilter.args['mask']
            self.mask_list = []

    def setup(self):
        """
        Create mask list.

        Consists of all tuples between which this filter accepts lines.
        """
        # get start and end of the mask and set a start_limit
        if not self.mask_source.start:
            raise SystemExit("Can't parse format of %s. Is this a log file or "
                             "system.profile collection?"
                             % self.mlogfilter.args['mask'])

        self.mask_half_td = timedelta(seconds=self.mlogfilter.args
                                      ['mask_size'] / 2)

        # load filter mask file
        logevent_list = list(self.mask_source)

        # define start and end of total mask
        self.mask_start = self.mask_source.start - self.mask_half_td
        self.mask_end = self.mask_source.end + self.mask_half_td

        # consider --mask-center
        if self.mlogfilter.args['mask_center'] in ['start', 'both']:
            if logevent_list[0].duration:
                self.mask_start -= timedelta(milliseconds=logevent_list[0]
                                             .duration)

        if self.mlogfilter.args['mask_center'] == 'start':
            if logevent_list[-1].duration:
                self.mask_end -= timedelta(milliseconds=logevent_list[-1]
                                           .duration)

        self.start_limit = self.mask_start

        # different center points
        if 'mask_center' in self.mlogfilter.args:
            if self.mlogfilter.args['mask_center'] in ['start', 'both']:
                starts = ([(le.datetime - timedelta(milliseconds=le.duration))
                          if le.duration is not None else le.datetime
                          for le in logevent_list if le.datetime])

            if self.mlogfilter.args['mask_center'] in ['end', 'both']:
                ends = [le.datetime for le in logevent_list if le.datetime]

            if self.mlogfilter.args['mask_center'] == 'start':
                event_list = sorted(starts)
            elif self.mlogfilter.args['mask_center'] == 'end':
                event_list = sorted(ends)
            elif self.mlogfilter.args['mask_center'] == 'both':
                event_list = sorted(zip(starts, ends))

        mask_list = []

        if len(event_list) == 0:
            return

        start_point = end_point = None

        for e in event_list:
            if start_point is None:
                start_point, end_point = self._pad_event(e)
                continue

            next_start = (e[0] if type(e) == tuple else e) - self.mask_half_td
            if next_start <= end_point:
                end_point = ((e[1] if type(e) == tuple else e) +
                             self.mask_half_td)
            else:
                mask_list.append((start_point, end_point))
                start_point, end_point = self._pad_event(e)

        if start_point:
            mask_list.append((start_point, end_point))

        self.mask_list = mask_list

    def _pad_event(self, event):
        if type(event) == tuple:
            start_point = event[0] - self.mask_half_td
            end_point = event[1] + self.mask_half_td
        else:
            start_point = event - self.mask_half_td
            end_point = event + self.mask_half_td

        return start_point, end_point

    def accept(self, logevent):
        """
        Process line.

        Overwrite BaseFilter.accept() and return True if the provided
        logevent should be accepted (causing output), or False if not.
        """
        dt = logevent.datetime
        if not dt:
            return False

        mask = next((mask for mask in self.mask_list
                     if mask[0] < dt and mask[1] > dt), None)

        return True if mask else False

    def skipRemaining(self):
        """
        Skip remaining lines.

        Overwrite BaseFilter.skipRemaining() and return True if all lines
        from here to the end of the file should be rejected (no output).
        """
        return self.mask_end_reached



from collections import defaultdict, namedtuple, Counter
from difflib import SequenceMatcher
import re

from mtools.util import OrderedDict
from mtools.util.grouping import Grouping
from mtools.util.print_table import print_table
from .base_section import BaseSection

ChunksTuple = namedtuple('ChunksTuple', [
    'time',
    'range',
    'movedFromTo',
    'namespace',
    'steps',
    'migrationStatus',
    'errorMessage'
])

SplitTuple = namedtuple('SplitTuple', [
    'time',
    'range',
    'namespace',
    'numSplits',
    'success',
    'timeTaken',
    'error'
])


class ShardingSection(BaseSection):
    """
    ShardingSection class.

    This section goes through the logfile and extracts any sharding related information
    """

    name = "sharding"

    def __init__(self, mloginfo):
        BaseSection.__init__(self, mloginfo)

        helptext = 'outputs sharding related information'
        self.mloginfo.argparser_sectiongroup.add_argument('--sharding',
                                                          action='store_true',
                                                          help=helptext)
        helptext = 'toggle output for sharding related errors/warnings'
        self.mloginfo.argparser_sectiongroup.add_argument('--errors',
                                                          action='store_true',
                                                          help=helptext)
        helptext = 'toggle output for shard chunk migration'
        self.mloginfo.argparser_sectiongroup.add_argument('--migrations',
                                                          action='store_true',
                                                          help=helptext)

    @property
    def active(self):
        """Return boolean if this section is active."""
        return(self.mloginfo.args['sharding'])

    def run(self):
        """Run this section and print out information."""
        logfile = self.mloginfo.logfile

        print("\nOverview:\n")

        if logfile.shards and logfile.csrs:

            if logfile.binary == "mongos":
                print("  The role of this node: (mongos)")
            elif logfile.repl_set in logfile.csrs[0]:
                print("  The role of this node: (CSRS)")
            else:
                print("  The role of this node: (shard)")

            print("  Shards:")
            for shard_name, replica_set in logfile.shards:
                print(f"    {shard_name}: {replica_set}")

            print("  CSRS:")
            name, replica_set = logfile.csrs
            print(f"    {name}: {replica_set}")
        else:
            print("  no sharding info found.")

        if logfile.start and logfile.end:
            progress_start = self.mloginfo._datetime_to_epoch(logfile.start)
            progress_total = (self.mloginfo._datetime_to_epoch(logfile.end) - progress_start)
        else:
            self.mloginfo.progress_bar_enabled = False

        if self.mloginfo.args['errors']:
            errorlines = defaultdict(lambda: 0)
            for i, logevent in enumerate(logfile):

                # update progress bar every 1000 lines
                if self.mloginfo.progress_bar_enabled and (i % 1000 == 0):
                    if logevent.datetime:
                        progress_curr = self.mloginfo._datetime_to_epoch(logevent.datetime)

                        if progress_total:
                            (self.mloginfo.update_progress(float(progress_curr - progress_start) /
                             progress_total))

                # All common error message is lower case so it can be compared to line content
                common_error_message_content = [
                    'failed to update the persisted chunk metadata for collection ... caused by',
                    'cannot accept new chunks because there are still ... deletes from '
                    'previous migration',
                    'local document ... has same _id as cloned remote document',
                    'local document ... has same _id as reloaded remote document',
                    'batch insertion failed'
                    '[rangedeleter] waiting for open cursors before removing range',
                    'will be scheduled after all possibly dependent queries finish',
                    '[collection range deleter] deferring deletion of '
                ]

                if (not any(keyword in logevent.line_str.lower() for keyword in
                            common_error_message_content)):
                    continue

                log_tokens = logevent.split_tokens[3:]

                for index, token in enumerate(log_tokens):
                    if re.match(r'\'(.+?)\'', token) or any(char.isdigit() for char in token):
                        log_tokens[index] = "..."

                error_log_line = ' '.join(log_tokens)
                error_log_line = re.sub(r' \S+\.\S+ ', ' XXX ', error_log_line)
                error_log_line = re.sub(r'\{.*\}', '...', error_log_line)

                if not errorlines.keys():
                    errorlines[error_log_line] += 1
                    continue

                similar_error = False
                for errorline in errorlines.keys():
                    similar_error = SequenceMatcher(None, errorline, error_log_line).ratio() >= 0.7
                    if similar_error:
                        errorlines[errorline] += 1
                        break

                if not similar_error:
                    errorlines[error_log_line] += 1

            # clear progress bar again
            if self.mloginfo.progress_bar_enabled:
                self.mloginfo.update_progress(1.0)

            print("Error Messages:\n")

            if not len(errorlines):
                print("  no error messages found.")
            else:
                for cl in sorted(errorlines, key=lambda x: errorlines[x], reverse=True):
                    print("%3i  %s" % (errorlines[cl], cl))
        else:
            print("\nto show sharding errors/warnings, run with --errors.")

        if self.mloginfo.args['migrations']:
            print("\nChunks Moved From This Shard:\n")
            self._print_chunk_migrations(logfile.chunks_moved_from, moved_from=True)

            print("\nChunks Moved To This Shard:\n")
            self._print_chunk_migrations(logfile.chunks_moved_to)

            print("\nChunk Split Statistics:\n")
            self._print_chunk_statistics()
        else:
            print("\nto show chunk migrations/splits, run with --migrations.")

        print("")

    def _print_chunk_statistics(self):
        """Prints the chunk split statistics in a table"""
        self.mloginfo.logfile.chunk_splits.reverse()

        chunk_split_groupings = Grouping(group_by=lambda x: (x.time.strftime("%Y-%m-%dT%H"),
                                                             x.namespace))

        for chunk_split in self.mloginfo.logfile.chunk_splits:
            time, split_range, namespace, numSplits, success, timeTaken, error = chunk_split
            split_tuple = SplitTuple(time=time,
                                     range=split_range,
                                     namespace=namespace,
                                     numSplits=numSplits,
                                     success=success,
                                     timeTaken=timeTaken,
                                     error=error)
            chunk_split_groupings.add(split_tuple)

        titles = ['  time (/hour)',
                  'namespace',
                  '# split-vectors issued',
                  'successful chunk splits',
                  'failed chunk splits']

        if len(chunk_split_groupings) == 0:
            print("  no chunk splits found.")
        else:
            table_rows = []
            for group, splits in chunk_split_groupings.items():

                time, namespace = group
                successful_count = 0
                total_number_vectors = 0
                split_succeeded_after = dict()
                failed_splits = dict()
                total_time_taken = 0
                for split in splits:
                    total_number_vectors += int(split.numSplits)
                    if (not split.success) and split.error:
                        count, timestamps = failed_splits.get(split.error, (0, list()))
                        count += 1
                        if split_succeeded_after.get(split.range, False):
                            timestamps.append(split.time.strftime("%H:%M:%S.%f")[:-3] +
                                              ' **WAS SUCCESSFUL AFTER**')
                        else:
                            timestamps.append(split.time.strftime("%H:%M:%S.%f")[:-3])
                        failed_splits[split.error] = (count, timestamps)
                    elif split.success:
                        split_succeeded_after[split.range] = True
                        successful_count += 1
                        total_time_taken += sum(int(ms) for ms in split.timeTaken)

                split_summary = OrderedDict()

                split_summary['time'] = f"  {time}"
                split_summary['namespace'] = namespace

                split_summary['numSplitVectors'] = f'{total_number_vectors} split vector(s)'
                msg = (f"{successful_count} chunk(s) splitted" +
                       f" | Total time spent: {total_time_taken}ms")
                split_summary['successfulSplits'] = msg

                failed_split = ""
                for error, info in failed_splits.items():
                    count, timestamps = info
                    if error == "Jumbo":
                        failed_split += (f'{count} chunk(s): ' +
                                         f'{timestamps} marked as {error}.')
                    else:
                        failed_split += (f'{count} chunk(s): {timestamps} ' +
                                         f'failed with "{error}". ')

                if len(failed_split):
                    split_summary['failedChunkSplits'] = failed_split
                else:
                    split_summary['failedChunkSplits'] = "no failed chunk splits."

                table_rows.append(split_summary)

            print_table(table_rows, titles)

    def _print_chunk_migrations(self, chunks, moved_from=False):
        """Prints the chunk migration statistics in a table depending on to/from flag"""
        verbose = self.mloginfo.args['verbose']
        chunks.reverse()

        if verbose:
            chunk_groupings = Grouping(group_by=lambda x: x.time)
        else:
            chunk_groupings = Grouping(group_by=lambda x: (x.time.strftime("%Y-%m-%dT%H"),
                                                           x.movedFromTo,
                                                           x.namespace))

        for chunk_moved in chunks:
            time, chunk_range, moved_to_from, namespace, steps, status, error_message = chunk_moved
            moved_tuple = ChunksTuple(time=time,
                                      range=chunk_range,
                                      movedFromTo=moved_to_from,
                                      namespace=namespace,
                                      steps=steps,
                                      migrationStatus=status,
                                      errorMessage=error_message)
            chunk_groupings.add(moved_tuple)

        move_to_from_title = 'to shard' if moved_from else 'from shard'
        if verbose:
            titles = ['  time', move_to_from_title,
                      'namespace',
                      'chunk migration status']
        else:
            titles = ['  time (/hour)', move_to_from_title,
                      'namespace',
                      '# chunks migrations attempted',
                      'successful chunk migrations',
                      'failed chunk migrations']

        if len(chunk_groupings) == 0:
            print("  no chunk migrations found.")
        else:
            table_rows = []
            for group, chunks in chunk_groupings.items():

                if verbose:
                    time = group.strftime("%Y-%m-%dT%H:%M:%S.%f")[:-3]
                    chunk = chunks[0]
                else:
                    time, moved_to_from, namespace = group
                    successful_count = 0
                    total_time_spent = 0
                    failed = dict()
                    succeeded_after = dict()
                    for chunk in chunks:
                        if chunk.migrationStatus == "success":
                            successful_count += 1
                            succeeded_after[chunk.range] = (True, chunk.time)
                            total_time_spent += sum(int(ms) for step, ms in chunk.steps)
                        else:
                            count, timestamps = failed.get(chunk.errorMessage, (0, list()))
                            count += 1
                            successful_after, timestamp = succeeded_after.get(chunk.range,
                                                                              (False, None))
                            if successful_after:
                                timestamp = timestamp.strftime("%H:%M:%S.%f")[:-3]
                                chunk_time = chunk.time.strftime("%H:%M:%S.%f")[:-3]
                                timestamps.append(chunk_time +
                                                  f' BECAME SUCCESSFUL AT: {timestamp}')
                            else:
                                timestamps.append(chunk.time.strftime("%H:%M:%S.%f")[:-3])
                            failed[chunk.errorMessage] = (count, timestamps)

                moved_chunks = OrderedDict()

                moved_chunks['time'] = f"  {time}"
                moved_chunks['movedFromTo'] = moved_to_from
                moved_chunks['namespace'] = namespace

                if verbose:
                    if chunk.migrationStatus == "success":
                        total_time_spent = sum(int(ms) for step, ms in chunk.steps)
                        msg = f"Successful | Total time spent {total_time_spent}ms"
                        step_breakdown = ', '.join(f"{step}: {ms}ms" for step, ms in chunk.steps)
                        moved_chunks['chunkMigrationStatus'] = msg + f" ({step_breakdown})"
                    else:
                        moved_chunks['chunkMigrationStatus'] = f"Failed with {chunk.errorMessage}"
                else:
                    moved_chunks['numberOfChunks'] = f'{len(chunks)} chunk(s)'
                    msg = (f"{successful_count} chunk(s) moved " +
                           f"| Total time spent: {total_time_spent}ms")
                    moved_chunks['successChunkMigrations'] = msg

                    failed_migrations = ""
                    for error, info in failed.items():
                        count, timestamps = info
                        failed_migrations += (f'{count} chunk(s): {timestamps} '
                                              f'failed with "{error}".')

                    if len(failed_migrations):
                        moved_chunks['failedChunkMigrations'] = failed_migrations
                    else:
                        moved_chunks['failedChunkMigrations'] = "no failed chunks."

                table_rows.append(moved_chunks)

            print_table(table_rows, titles)
            if not verbose:
                print("\nto show individual chunk migration, run with --verbose.")



#!/usr/bin/env python3

import json
import re

import sys


def _decode_pattern_list(data):
    rv = []
    contains_dict = False
    for item in data:
        if isinstance(item, list):
            item = _decode_pattern_list(item)
        elif isinstance(item, dict):
            item = _decode_pattern_dict(item)
            contains_dict = True
        rv.append(item)

    # avoid sorting if any element in the list is a dict
    if not contains_dict:
        rv = sorted(rv)

    return rv


def _decode_pattern_dict(data):
    rv = {}
    for key, value in data.items():
        if isinstance(key, bytes):
            key = key.encode('utf-8')
        if isinstance(key, str):
            if key in ['$in', '$gt', '$gte', '$lt', '$lte', '$exists']:
                return 1
            if key == '$nin':
                value = 1
            if key in ['query', '$query']:
                try:
                    # Try to decode value as a dictionary; this will fail if
                    # there happens to be a field called "query"
                    return _decode_pattern_dict(value)
                except:
                    return value

        if isinstance(value, list):
            value = _decode_pattern_list(value)
        elif isinstance(value, dict):
            value = _decode_pattern_dict(value)
        else:
            value = 1
        rv[key] = value
    return rv


def shell2json(s):
    """Convert shell syntax to json."""
    replace = {
        r'BinData\(.+?\)': '1',
        r'(new )?Date\(.+?\)': '1',
        r'Timestamp\(.+?\)': '1',
        r'ObjectId\(.+?\)': '1',
        r'DBRef\(.+?\)': '1',
        r'undefined': '1',
        r'MinKey': '1',
        r'MaxKey': '1',
        r'NumberLong\(.+?\)': '1',
        r'/.+?/\w*': '1'
    }

    for key, value in replace.items():
        s = re.sub(key, value, s)

    return s


def json2pattern(s, debug = False):
    """
    Convert JSON format to a query pattern.

    Includes even mongo shell notation without quoted key names.

    Pass debug = True to print additional info on each step of processing chain
    """
    saved_s = s

    if debug : print ("\n=======================\n", saved_s, file=sys.stderr)

    # make valid JSON by wrapping field names in quotes
    s, _ = re.subn(r'([{,])\s*([^,{\s\'"]+)\s*:', ' \\1 "\\2" : ', s)
    if debug : print (s, file=sys.stderr) 

    # handle shell values that are not valid JSON
    s = shell2json(s)
    if debug : print (s, file=sys.stderr)
    
    # convert to 1 where possible, to get rid of things like new Date(...)
    s, _ = re.subn(r'([:,\[])\s*([^{}\[\]"]+?)\s*([,}\]])', '\\1 1 \\3', s)
    if debug : print (s, file=sys.stderr)


    # replace list values by 1. Not the '$in/$nin' lists, but the like of: {..., "attrib" : ["val1", "val2", "3"],...}
    # updated regex, using positive lookahead and lookbehind to check for a " (quote) 
    # right after '['  and before ']' to correctly handle cases where a ']' is part of the value and 
    # also cases where list values are url's "nnn://aaa.bbb"  will correctly be simplified to '1'
    s, _ = re.subn(r'("\S+"\s*:\s*\[\s*(?=\"))(.+)((?<=\")\s*\]\s*[,}])', '\\1 1 \\3', s)

    if debug : print (s, file=sys.stderr)

    # now convert to dictionary, converting unicode to ascii
    doc = {}
    try:
        doc = json.loads(s, object_hook=_decode_pattern_dict)
    except Exception as err:
        if debug:
            ## print some context info and return without any extracted query data..
            msg = f'''json2pattern():json.loads Exception:\n  Error: {err} : {sys.exc_info()[0]}\n  saved_s: ({saved_s})\n  s: ({s})\n'''
            print(msg, file=sys.stderr)
        return None
    except:
        print (f'''json2pattern():json.loads Unexpected error: saved_s: ({saved_s}) sys.exc_info():{sys.exc_info()[0]}''' )
        raise


    try:
        return json.dumps(doc, sort_keys=True, separators=(', ', ': '), ensure_ascii=False)
    except Exception as err:
        ## print some context info and return without any extracted query data..
        if debug:
            msg = f'''json2pattern():json.dumps Exception:\n  Error: {sys.exc_info()[0]} : {err}\n  saved_s: ({saved_s})\n  doc: ({doc})\n'''
            print(msg, file=sys.stderr)
        return None
    except:
        print(f'''json2pattern():json.dumps Unexpected error: saved_s: ({saved_s}) sys.exc_info():{sys.exc_info()[0]}''')
        raise


if __name__ == '__main__':

    # define as True to get debug output of regex processing printed to stderr
    debug = False

    tests = { 
        '{d: {$gt: 2, $lt: 4}, b: {$gte: 3}, c: {$nin: [1, "foo", "bar"]}, "$or": [{a:"1uno"}, {b:"1uno"}] }'                : '{"$or": [{"a": 1}, {"b": 1}], "b": 1, "c": {"$nin": 1}, "d": 1}',
        '{a: {$gt: 2, $lt: 4}, "b": {$nin: [1, 2, 3]}, "$or": [{a:1}, {b:1}] }'                                              : '{"$or": [{"a": 1}, {"b": 1}], "a": 1, "b": {"$nin": 1}}', 
        "{a: {$gt: 2, $lt: 4}, b: {$in: [ ObjectId('1234564863acd10e5cbf5f6e'), ObjectId('1234564863acd10e5cbf5f7e') ] } }"  : '{"a": 1, "b": 1}', 
        "{ sk: -1182239108, _id: { $in: [ ObjectId('1234564863acd10e5cbf5f6e'), ObjectId('1234564863acd10e5cbf5f7e') ] } }"  : '{"_id": 1, "sk": 1}', 
        '{ a: 1, b: { c: 2, d: "text" }, e: "more test" }'                                                                   : '{"a": 1, "b": {"c": 1, "d": 1}, "e": 1}', 
        '{ _id: ObjectId(\'528556616dde23324f233168\'), config: { _id: 2, host: "localhost:27017" }, ns: "local.oplog.rs" }' : '{"_id": 1, "config": {"_id": 1, "host": 1}, "ns": 1}',

        # 20191231 - bugre - issue#764 - adding some more test cases.. based on our mongodb logs (mongod 4.0.3)
        r'{_id: ObjectId(\'528556616dde23324f233168\'), curList: [ "€", "XYZ", "Krown"], allowedSnacks: 1000 }'              : '{"_id": 1, "allowedSnacks": 1, "curList": [1]}', 
        r'{_id: "test", curList: [ "1onum]pas", "ab\]c" ] }'                                                                 : '{"_id": 1, "curList": [1]}',
        r'{ $and: [ { mode: ObjectId(\'5aafd085edb85e0dc09dd985\') }, { _id: { $ne: ObjectId(\'5e015519877718752d63dd9c\') } }, ' 
            '{ snack: { $in: [ "BLA", "RUN", "BLE" ] } }, { $or: [ { $and: [ { kind: "Solar" }, { wind: true }, '
            '{ beginTime: { $gte: new Date(1577134729858) } } ] }, { $and: [ { kind: "event" }, { endTime: { $gte: new Date(1577739529858) } } ] } ] } ] }'  : 
                '{"$and": [{"mode": 1}, {"_id": {"$ne": 1}}, {"snack": 1}, {"$or": [{"$and": [{"kind": 1}, {"wind": 1}, {"beginTime": 1}]}, {"$and": [{"kind": 1}, {"endTime": 1}]}]}]}',
        
        # @niccottrell use case and 2nd one extrapolating the 1st one. 
        r'{ urls: { $all: [ "https://surtronic.info/" ] } }'                      : '{"urls": {"$all": [1]}}',
        r'{ urls: { $all: [ "https://surtronic.info/", "http://url2.com" ] } }'   : '{"urls": {"$all": [1]}}'
    }

    for k,v in tests.items():
        r = json2pattern(k, debug)
        if ( r == v ):
            if debug :
                print(f'''OK...: {k}\n  Expect: {v}\n  Output: {r}\n\n''')
            else:
                print(f'''OK: {k}''')

        else:
            print(f'''\nERROR **: {k}\n  Expect: {v}\n  Output: {r}\n\n''')
